import csv
import typing as tp
from pathlib import Path

from .model import SearchRequest

if tp.TYPE_CHECKING:
    from patent_client.util.manager import ManagerConfig

field_file = Path(__file__).parent / "query_fields.csv"

field_index = dict()

with field_file.open("r") as file:
    reader = csv.DictReader(file)
    for row in reader:
        record = {
            "field_name": row["field_name"],
            "type": row["type"],
        }
        field_index[row["field_name"]] = record
        field_index[row["underscore_case"]] = record
        field_index[row["simple_name"]] = record


def create_post_search_obj(
    config: "ManagerConfig", fields: tp.Optional[tp.List[str]] = None
) -> "SearchRequest":
    query_obj = dict(
        pagination=dict(offset=0, limit=100),
        sort=list(),
    )
    q_str = list()
    range_filters = dict()
    for field, value in config.filter.items():
        if field.endswith("_gte") or field.endswith("_lte"):
            field_name, tail = field.rsplit("_", 1)
            if field_name not in field_index:
                raise ValueError(f"Unknown field: {field_name}")
            field_name = field_index[field_name]["field_name"]
            if field.endswith("_gte"):
                range_filters[field_name] = {"valueFrom": value[0]}
            elif field.endswith("_lte"):
                range_filters[field_name] = {"valueTo": value[0]}

        else:
            if field not in field_index:
                raise ValueError(f"Unknown field: {field}")
            field_data = field_index[field]
            if field_data["type"] == "search":
                q_str.append(f"{field_data['field_name']}:({' OR '.join(value)})")
            else:
                q_str.append(f"{field_data['field_name']}=({' OR '.join(value)})")

    for sort_field, direction in config.order_by:
        query_obj["sort"].append(
            dict(field=field_index[sort_field]["field_name"], direction=direction)
        )
    query_obj["q"] = " AND ".join(q_str)
    query_obj["rangeFilters"] = [{"field": k, **v} for k, v in range_filters.items()]
    query_obj["pagination"]["offset"] = config.offset
    query_obj["pagination"]["limit"] = config.limit or 25
    query_obj["fields"] = (
        fields
        if fields is not None
        else [
            "applicationNumberText",
        ]
    )
    return SearchRequest(**query_obj)



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *       Source File: patent_client/_async/uspto/global_dossier/model.py        *
# ********************************************************************************

import datetime
import typing as tp
from pathlib import Path

from pydantic import BeforeValidator, ConfigDict, Field, model_validator
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated

from patent_client.util.pydantic_util import BaseModel

if tp.TYPE_CHECKING:
    from ..assignment.model import Assignment
    from ..peds.model import USApplication
    from ..public_search.model import Patent, PublishedApplication
MDYDate = Annotated[
    datetime.date,
    BeforeValidator(lambda x: datetime.datetime.strptime(x, "%m/%d/%Y").date()),
]
OptionalStrList = Annotated[
    tp.List[str], BeforeValidator(lambda x: x if isinstance(x, list) else list())
]


class GlobalDossierBaseModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        str_strip_whitespace=True,
    )


class GlobalDossierPublication(GlobalDossierBaseModel):
    pub_country: tp.Optional[str] = None
    pub_num: tp.Optional[str] = None
    kind_code: tp.Optional[str] = None
    pub_date: tp.Optional[MDYDate] = Field(alias="pubDateStr")


class GlobalDossierPriorityClaim(GlobalDossierBaseModel):
    country: tp.Optional[str] = None
    doc_number: tp.Optional[str] = None
    kind_code: tp.Optional[str] = None


class GlobalDossierDocumentNumber(GlobalDossierBaseModel):
    country: tp.Optional[str] = None
    doc_number: tp.Optional[str] = None
    format: tp.Optional[str] = None
    date: tp.Optional[MDYDate] = None
    kind_code: tp.Optional[str] = None


class GlobalDossierApplication(GlobalDossierBaseModel):
    app_num: tp.Optional[str] = None
    app_date: tp.Optional[MDYDate] = Field(alias="appDateStr")
    country_code: tp.Optional[str] = None
    kind_code: tp.Optional[str] = None
    doc_num: tp.Optional["GlobalDossierDocumentNumber"] = None
    title: tp.Optional[str] = None
    applicant_names: OptionalStrList = Field(default_factory=list)
    ip_5: tp.Optional[bool] = Field(alias="ip5")
    priority_claim_list: "tp.List[GlobalDossierPriorityClaim]" = Field(default_factory=list)
    pub_list: "tp.List[GlobalDossierPublication]" = Field(default_factory=list)

    def __repr__(self):
        return f"GlobalDossierApplication(app_num={self.app_num}, country_code={self.country_code})"

    @property
    def document_list(self) -> "DocumentList":
        return self._get_model(".model.DocumentList").objects.get(
            self.country_code, self.app_num, self.kind_code
        )

    @property
    def documents(self) -> list["Document"]:
        return (
            self._get_model(".model.DocumentList").objects.get(
                self.country_code, self.app_num, self.kind_code
            )
        ).docs

    @property
    def office_actions(self) -> list["Document"]:
        return (
            self._get_model(".model.DocumentList").objects.get(
                self.country_code, self.app_num, self.kind_code
            )
        ).office_action_docs

    @property
    def us_application(self) -> "USApplication":
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        return self._get_model("..peds.model.USApplication").objects.get(self.app_num)

    @property
    def us_publication(self) -> "PublishedApplication":
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        pub = list(p for p in self.pub_list if p.kind_code == "A1")
        if len(pub) > 1:
            raise ValueError("More than one US publication for application!")
        return self._get_model("..public_search.model.PublishedApplication").objects.get(
            pub[0].pub_num
        )

    @property
    def us_patent(self) -> "Patent":
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        pat = list(p for p in self.pub_list if p.kind_code in ("A", "B1", "B2"))
        if len(pat) > 1:
            raise ValueError("More than one US patent for application!")
        return self._get_model("..public_search.model.Patent").objects.get(pat[0].pub_num)

    @property
    def us_assignments(self) -> list["Assignment"]:
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        return self._get_model("..assignment.model.Assignment").objects.filter(appl_id=self.app_num)


TextBool = Annotated[bool, BeforeValidator(lambda x: x == "true")]


class GlobalDossier(GlobalDossierBaseModel):
    country: tp.Optional[str] = None
    internal: tp.Optional[TextBool] = None
    corr_app_num: tp.Optional[str] = None
    id: tp.Optional[str] = None
    type: tp.Optional[str] = None
    applications: tp.List[GlobalDossierApplication] = Field(default_factory=list, alias="list")

    def __repr__(self):
        return f"GlobalDossier(id={self.id}, type={self.type}, country={self.country})"


class Document(GlobalDossierBaseModel):
    doc_number: tp.Optional[str] = None
    country: tp.Optional[str] = None
    doc_code: tp.Optional[str] = None
    doc_desc: tp.Optional[str] = None
    doc_id: tp.Optional[str] = None
    date: tp.Optional[MDYDate] = Field(alias="legalDateStr")
    doc_format: tp.Optional[str] = None
    pages: tp.Optional[int] = Field(alias="numberOfPages")
    doc_code_desc: tp.Optional[str] = None
    doc_group_code: tp.Optional[str] = None
    shareable: tp.Optional[bool] = None

    def download(self, filename="", path="."):
        from .manager import global_dossier_api

        out_path = Path(path).expanduser() / (
            f'{self.date.isoformat()} - {self.doc_desc.replace("/", "_")}.pdf'
            if not filename
            else filename
        )
        return global_dossier_api.get_document(
            self.country, self.doc_number, self.doc_id, out_path=out_path
        )


class DocumentList(GlobalDossierBaseModel):
    title: tp.Optional[str] = None
    doc_number: tp.Optional[str] = None
    country: tp.Optional[str] = None
    message: tp.Optional[str] = None
    applicant_names: "tp.List[str]" = Field(default_factory=list)
    docs: tp.List[Document] = Field(default_factory=list)
    office_action_docs: tp.List[Document] = Field(default_factory=list)

    @model_validator(mode="before")
    @classmethod
    def annotate_docs_and_office_actions(cls, values):
        update_dict = {
            "country": values["country"],
            "docNumber": values["docNumber"],
        }
        for doc in values["docs"]:
            doc.update(update_dict)
        for doc in values["officeActionDocs"]:
            doc.update(update_dict)
        return values



import logging
from io import BytesIO

from yankee.data import AttrDict

from ..session import session
from .model.biblio import BiblioResult
from .model.fulltext import Claims, Description
from .model.images import Images
from .model.search import Search

logger = logging.getLogger(__name__)


class PublishedBiblioApi:
    @classmethod
    async def get_constituents(
        cls, number, doc_type="publication", format="docdb", constituents=("biblio",)
    ):
        """Published Data Constituents API
        number: document number to search
        doc_type: document type (application / publication)
        format: document number format (original / docdb / epodoc)
        constituents: what data to retrieve. Can be combined. (biblio / abstract / full-cycle)

        """
        base_url = (
            f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/"
        )
        if isinstance(constituents, str):
            constituents = (constituents,)
        url = base_url + ",".join(constituents)
        response = await session.get(url)
        return response.text

    @classmethod
    async def get_biblio(cls, number, doc_type="publication", format="docdb") -> "BiblioResult":
        text = await cls.get_constituents(number, doc_type, format, constituents="biblio")
        return BiblioResult.model_validate(text)

    @classmethod
    async def get_abstract(cls, number, doc_type="publication", format="docdb"):
        return await cls.get_constituents(number, doc_type, format, constituents="abstract")

    @classmethod
    async def get_full_cycle(cls, number, doc_type="publication", format="docdb"):
        return await cls.get_constituents(number, doc_type, format, constituents="full-cycle")


class PublishedFulltextApi:
    fulltext_jurisdictions = "EP, WO, AT, BE, BG, CA, CH, CY, CZ, DK, EE, ES, FR, GB, GR, HR, IE, IT, LT, LU, MC, MD, ME, NO, PL, PT, RO, RS, SE, SK".split(
        ", "
    )

    @classmethod
    async def get_fulltext_result(
        cls, number, doc_type="publication", format="docdb", inquiry="fulltext"
    ):
        """Published Fulltext API
        number: document number to search
        doc_type: document type (application / publication)
        format: document number format (original / docdb / epodoc)
        inquiry: what data to retrieve. Can be combined. (fulltext / description / claims)

        """
        url = f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/{inquiry}"
        if number[:2] not in cls.fulltext_jurisdictions:
            raise ValueError(
                f"Fulltext Is Not Available For Country Code {number[:2]}. Fulltext is only available in {', '.join(cls.fulltext_jurisdictions)}"
            )
        response = await session.get(url)
        response.raise_for_status
        return response.text

    @classmethod
    async def get_description(cls, number, doc_type="publication", format="docdb"):
        text = await cls.get_fulltext_result(
            number, doc_type="publication", format="docdb", inquiry="description"
        )
        return Description.model_validate(text)

    @classmethod
    async def get_claims(cls, number, doc_type="publication", format="docdb") -> "Claims":
        text = await cls.get_fulltext_result(
            number, doc_type="publication", format="docdb", inquiry="claims"
        )
        return Claims.model_validate(text)


class PublishedSearchApi:
    @classmethod
    async def search(cls, query, start=1, end=100):
        base_url = "http://ops.epo.org/3.2/rest-services/published-data/search"
        range = f"{start}-{end}"
        logger.debug(f"OPS Search Endpoint - Query: {query}\nRange: {start}-{end}")
        response = await session.get(base_url, params={"Range": range, "q": query})
        if response.status_code == 404:
            return AttrDict.convert(
                {
                    "query": "query",
                    "num_results": 0,
                    "begin": start,
                    "end": end,
                    "results": list(),
                }
            )

        return Search.model_validate(response.text)


class PublishedImagesApi:
    @classmethod
    async def get_images(cls, number, doc_type="publication", format="docdb"):
        base_url = f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/images"
        response = await session.get(base_url)
        return Images.model_validate(response.text)

    @classmethod
    async def get_page_image(
        cls, country, number, kind, image_type, page_number, image_format="pdf"
    ):
        response = await session.get(
            f"https://ops.epo.org/3.2/rest-services/published-data/images/{country}/{number}/{kind}/{image_type}.{image_format}",
            params={"Range": page_number},
            extensions={"force_cache": True},
        )
        return BytesIO(response.content)

    @classmethod
    async def get_page_image_from_link(cls, link, page_number, image_format="pdf"):
        response = await session.get(
            f"https://ops.epo.org/3.2/rest-services/{link}.{image_format}",
            params={"Range": page_number},
            extensions={"force_cache": True},
        )
        return BytesIO(response.content)


class PublishedApi:
    biblio = PublishedBiblioApi
    fulltext = PublishedFulltextApi
    search = PublishedSearchApi
    images = PublishedImagesApi



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *   Source File: patent_client/_async/uspto/public_search/model/__init__.py    *
# ********************************************************************************

from .biblio import PublicSearchBiblio, PublicSearchBiblioPage
from .document import PublicSearchDocument


class PatentBiblio(PublicSearchBiblio):
    pass


class Patent(PublicSearchDocument):
    pass


class PublishedApplicationBiblio(PublicSearchBiblio):
    pass


class PublishedApplication(PublicSearchDocument):
    pass


__all__ = [
    "PatentBiblio",
    "Patent",
    "PublishedApplicationBiblio",
    "PublishedApplication",
    "PublicSearchBiblio",
    "PublicSearchBiblioPage",
    "PublicSearchDocument",
    "PublicSearchDocument",
]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *         Source File: patent_client/_async/epo/ops/family/manager.py          *
# ********************************************************************************

from patent_client.util.manager import Manager

from .api import FamilyAsyncApi
from .model import Family
from .schema import FamilySchema


class FamilyManager(Manager[Family]):
    __schema__ = FamilySchema

    def get(self, doc_number):
        return FamilyAsyncApi.get_family(doc_number, doc_type="publication", format="docdb")



from yankee.data import ListCollection

from patent_client.util.manager import AsyncManager

from .api import LegalApi
from .model import LegalEvent
from .schema import LegalSchema


class LegalManager(AsyncManager):
    __schema__ = LegalSchema

    async def get(
        self, doc_number, doc_type="publication", format="docdb"
    ) -> ListCollection[LegalEvent]:
        return self.__schema__.load(await LegalApi.get_legal(doc_number, doc_type, format)).events



import calendar
import datetime
import typing as tp

from patent_client.util.manager import AsyncManager

from .api import BulkDataApi

if tp.TYPE_CHECKING:
    from .model import File, Product


def date_ranges(start_date: datetime.date, end_date: datetime.date):
    # First range: start date to end of month
    end_of_month = datetime.datetime(
        start_date.year,
        start_date.month,
        calendar.monthrange(start_date.year, start_date.month)[1],
    )
    yield (start_date, end_of_month.date())

    # Full months between start and end date
    current_month = start_date.replace(day=1) + datetime.timedelta(days=32)
    while current_month.replace(day=1) < end_date.replace(day=1):
        last_day_of_month = current_month.replace(
            day=calendar.monthrange(current_month.year, current_month.month)[1]
        )
        yield (current_month.replace(day=1), last_day_of_month)
        current_month += datetime.timedelta(days=32)

    # Last range: start of month to end date
    yield (end_date.replace(day=1), end_date)


class ProductManager(AsyncManager):
    async def filter_by_latest(self) -> tp.AsyncIterator["Product"]:
        """Returns all products with Latest Files"""
        result = await BulkDataApi.get_latest()
        for product in result:
            yield product

    async def get_by_short_name(self, short_name) -> "Product":
        result = await BulkDataApi.get_by_short_name(short_name)
        return result

    async def filter_by_name(self, short_name) -> tp.AsyncIterator["Product"]:
        result = await BulkDataApi.get_by_name(short_name)
        for product in result:
            yield product


class FileManager(AsyncManager):
    async def filter_by_short_name(
        self, short_name, from_date=None, to_date=None
    ) -> tp.AsyncIterator["File"]:
        if from_date is not None:
            from_date = (
                from_date
                if isinstance(from_date, datetime.date)
                else datetime.date.fromisoformat(from_date)
            )
        if to_date is not None:
            to_date = (
                to_date
                if isinstance(to_date, datetime.date)
                else datetime.date.fromisoformat(to_date)
            )
        if from_date is None or to_date is None:
            product = await BulkDataApi.get_by_short_name(short_name)
            from_date = from_date or product.from_date
            to_date = to_date or product.to_date
        for start_date, end_date in date_ranges(from_date, to_date):
            chunk = await BulkDataApi.get_by_short_name(
                short_name, from_date=start_date, to_date=end_date
            )
            if chunk.files:
                for file in chunk.files:
                    yield file



from .biblio import PublicSearchBiblio, PublicSearchBiblioPage
from .document import PublicSearchDocument


class PatentBiblio(PublicSearchBiblio):
    pass


class Patent(PublicSearchDocument):
    pass


class PublishedApplicationBiblio(PublicSearchBiblio):
    pass


class PublishedApplication(PublicSearchDocument):
    pass


__all__ = [
    "PatentBiblio",
    "Patent",
    "PublishedApplicationBiblio",
    "PublishedApplication",
    "PublicSearchBiblio",
    "PublicSearchBiblioPage",
    "PublicSearchDocument",
    "PublicSearchDocument",
]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *             Source File: patent_client/_async/uspto/peds/api.py              *
# ********************************************************************************

import datetime
import logging
from typing import Dict, List, Optional

from httpx._exceptions import HTTPStatusError

from patent_client._sync.http_client import PatentClientSession

from .model import Document, PedsPage

logger = logging.getLogger(__name__)
type_map = {
    "string": str,
    "date": datetime.datetime,
    "text_general": str,
    "int": int,
    "text_ws": str,
}
client = PatentClientSession()


class PedsDownException(Exception):
    pass


class PatentExaminationDataSystemApi:
    base_url = "https://ped.uspto.gov/api"
    search_fields: Dict = dict()

    def is_online(self) -> bool:
        response = client.get(
            "https://ped.uspto.gov/api/search-fields",
            extensions={"cache_disabled": True},
        )
        if response.status_code == 200:
            return True, ""
        elif "requested resource is not available" in response.text:
            return False, "Patent Examination Data is Offline - this is a USPTO problem"
        elif "attempt failed or the origin closed the connection" in response.text:
            return (
                False,
                "The Patent Examination Data API is Broken! - this is a USPTO problem",
            )
        else:
            return False, "There is a USPTO problem"

    def check_response(self, response):
        try:
            response.raise_for_status()
        except HTTPStatusError as e:
            alive, reason = self.is_online()
            raise e if alive else PedsDownException(reason)

    @classmethod
    def get_search_fields(cls) -> dict:
        if hasattr(cls, "search_fields"):
            return cls.search_fields
        response = client.get("https://ped.uspto.gov/api/search-fields")
        cls().check_response(response)
        search_fields = response.json()
        cls.search_fields = {k: type_map[v] for k, v in search_fields.items()}
        return cls.search_fields

    def create_query(
        self,
        query: str,
        query_fields: Optional[str] = None,
        default_field: Optional[str] = "patentTitle",
        facet: bool = False,
        return_fields: Optional[List[str]] = None,
        filter_query: Optional[List[str]] = None,
        minimum_match: str = "100%",
        sort: Optional[str] = "applId asc",
        start: int = 0,
        rows: Optional[int] = None,
    ) -> "PedsPage":
        """
        Args:
            query (str): The query to be issued to PEDS
            query_fields (str): A list of fields to be queried
            default_field (str, optional): Default field to search. Defaults to None.
            facet (str, optional): Whether to enabled SOLR faceting. Defaults to "false".
            return_fields (str, optional): Specifies which fields should be returned. Defaults to "*".
            filter_query (str, optional)
        Returns:
            _type_: _description_
        """
        if query_fields is None:
            qf = "appEarlyPubNumber applId appLocation appType appStatus_txt appConfrNumber appCustNumber appGrpArtNumber appCls appSubCls appEntityStatus_txt patentNumber patentTitle inventorName firstNamedApplicant appExamName appExamPrefrdName appAttrDockNumber appPCTNumber appIntlPubNumber wipoEarlyPubNumber pctAppType firstInventorFile appClsSubCls rankAndInventorsList".split(
                " "
            )
        params: Dict[str, object] = {
            "df": default_field,
            "qf": " ".join(qf),
            "fl": " ".join(return_fields) if return_fields else "*",
            "fq": filter_query if filter_query else list(),
            "searchText": query,
            "sort": sort,
            "facet": "true" if facet else "false",
            "mm": minimum_match,
            "start": str(start),
        }
        if rows:
            params["rows"] = rows
        url = "https://ped.uspto.gov/api/queries"
        logger.debug(f"POST {url}\n{params}")
        response = client.post(
            url,
            json=params,
            headers={"Accept": "application/json"},
        )
        self.check_response(response)
        return PedsPage.model_validate(response.json())

    def get_documents(self, appl_id: str) -> List[Document]:
        url = f"https://ped.uspto.gov/api/queries/cms/public/{appl_id}"
        response = client.get(url)
        self.check_response(response)
        return [Document.model_validate(d) for d in response.json()]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *              Source File: patent_client/_async/uspto/odp/api.py              *
# ********************************************************************************

import typing as tp
from urllib.parse import quote

from patent_client import SETTINGS

from ...http_client import PatentClientSession
from .model import (
    Assignment,
    Continuity,
    CustomerNumber,
    Document,
    ForeignPriority,
    SearchGetRequest,
    SearchRequest,
    TermAdjustment,
    Transaction,
    USApplication,
    USApplicationBiblio,
)
from .util import prune


def urlescape(s: str) -> str:
    return quote(s, safe="")


class ODPApi:
    base_url = "https://beta-api.uspto.gov"

    def __init__(self):
        if SETTINGS.odp_api_key is None:
            raise ValueError("ODP API key is not set")
        self.client = PatentClientSession(headers={"X-API-KEY": SETTINGS.odp_api_key})

    def post_search(self, search_request: SearchRequest = SearchRequest()) -> tp.Dict:
        url = self.base_url + "/api/v1/patent/applications/search"
        search_data = prune(search_request.model_dump())
        response = self.client.post(url, json=search_data, headers={"accept": "application/json"})
        if response.status_code == 404 and "No matching records found" in response.text:
            return {
                "count": 0,
                "patentBag": [],
                "requestIdentifier": response.json()["requestIdentifier"],
            }
        response.raise_for_status()
        return response.json()

    def get_search(self, search_request: SearchGetRequest = SearchGetRequest()) -> tp.Dict:
        """Patent application search by supplying query parameters
        Query parameters are optional. When no query parameters supplied, top 25 applications are returned"""
        url = self.base_url + "/api/v1/patent/applications/search"
        search_data = prune(search_request.model_dump())
        response = self.client.get(url, params=search_data)
        if response.status_code == 404 and "No matching records found" in response.text:
            return {
                "count": 0,
                "patentBag": [],
                "requestIdentifier": response.json()["requestIdentifier"],
            }
        response.raise_for_status()
        return response.json()

    # Data Attributes
    def get_application_data(self, application_id: str) -> USApplication:
        """Patent application data by application id"""
        url = self.base_url + f"/api/v1/patent/applications/{urlescape(application_id)}"
        response = self.client.get(url)
        response.raise_for_status()
        return USApplication(**response.json()["patentBag"][0])

    def get_application_biblio_data(self, application_id: str) -> USApplicationBiblio:
        """Patent application basic data by application id"""
        url = (
            self.base_url
            + f"/api/v1/patent/applications/{urlescape(application_id)}/application-data"
        )
        response = self.client.get(url)
        response.raise_for_status()
        return USApplicationBiblio(**response.json()["patentBag"][0])

    def get_patent_term_adjustment_data(self, application_id: str) -> TermAdjustment:
        """Patent application term adjustment data by application id"""
        url = self.base_url + f"/api/v1/patent/applications/{urlescape(application_id)}/adjustment"
        response = self.client.get(url)
        response.raise_for_status()
        return TermAdjustment(**response.json()["patentBag"][0]["patentTermAdjustmentData"])

    def get_assignments(self, application_id: str) -> tp.List[Assignment]:
        """Patent application term adjustment data by application id"""
        url = self.base_url + f"/api/v1/patent/applications/{urlescape(application_id)}/assignment"
        response = self.client.get(url)
        response.raise_for_status()
        data = response.json()["patentBag"][0]["assignmentBag"]
        return [Assignment(**assignment) for assignment in data]

    def get_attorney_data(self, application_id: str) -> CustomerNumber:
        """Patent application attorney data by application id"""
        url = self.base_url + f"/api/v1/patent/applications/{urlescape(application_id)}/attorney"
        response = self.client.get(url)
        response.raise_for_status()
        return CustomerNumber(**response.json()["patentBag"][0]["recordAttorney"])

    def get_continuity_data(self, application_id: str) -> Continuity:
        """Patent application continuity data by application id"""
        url = self.base_url + f"/api/v1/patent/applications/{urlescape(application_id)}/continuity"
        response = self.client.get(url)
        response.raise_for_status()
        return Continuity(**response.json())

    def get_foreign_priority_data(self, application_id: str) -> tp.List[ForeignPriority]:
        """Patent application foreign priority data by application id"""
        url = (
            self.base_url
            + f"/api/v1/patent/applications/{urlescape(application_id)}/foreign-priority"
        )
        response = self.client.get(url)
        response.raise_for_status()
        return [
            ForeignPriority(**foreign_priority)
            for foreign_priority in response.json()["patentBag"][0]["foreignPriorityBag"]
        ]

    def get_transactions(self, application_id: str) -> tp.List[Transaction]:
        """Patent application transactions by application id"""
        url = (
            self.base_url + f"/api/v1/patent/applications/{urlescape(application_id)}/transactions"
        )
        response = self.client.get(url)
        response.raise_for_status()
        return [
            Transaction(**transaction)
            for transaction in response.json()["patentBag"][0]["transactionContentBag"]
        ]

    def get_documents(self, application_id: str) -> tp.List[Document]:
        """Patent application documents by application id"""
        url = self.base_url + f"/api/v1/patent/applications/{urlescape(application_id)}/documents"
        response = self.client.get(url)
        response.raise_for_status()
        return [Document(**document) for document in response.json()["documentBag"]]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *           Source File: patent_client/_async/epo/ops/family/api.py            *
# ********************************************************************************

from ..session import session
from .model import Family


class FamilyAsyncApi:
    @classmethod
    def get_family(cls, number, doc_type="publication", format="docdb"):
        url = f"http://ops.epo.org/3.2/rest-services/family/{doc_type}/{format}/{number}"
        response = session.get(url)
        return Family.model_validate(response.text)



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *        Source File: patent_client/_async/uspto/public_search/util.py         *
# ********************************************************************************

import re

import lxml.html as ETH

newline_re = re.compile(r"<br />\s*")
bad_break_re = re.compile(r"<br />\s+")


def html_to_text(html):
    html = newline_re.sub("\n\n", html)
    # html = bad_break_re.sub(" ", html)
    return "".join(ETH.fromstring(html).itertext())



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *         Source File: patent_client/_async/uspto/ptab/manager_test.py         *
# ********************************************************************************

from .model import PtabDecision, PtabDocument, PtabProceeding


class TestPtabProceeding:
    def test_get_by_proceeding_number(self):
        result = PtabProceeding.objects.get("IPR2016-00831")
        assert result.respondent_patent_number == "6162705"

    def test_get_by_patent_number(self):
        result = PtabProceeding.objects.get(patent_number="6103599")
        assert result.proceeding_number == "IPR2016-00833"

    def test_get_by_application_number(self):
        result = PtabProceeding.objects.get(appl_id="09089931")
        assert result.proceeding_number == "IPR2016-00833"

    def test_filter_by_party(self):
        result = PtabProceeding.objects.filter(party_name="Apple")
        assert result.count() >= 400

    def test_filter_with_limit(self):
        result = PtabProceeding.objects.filter(party_name="Apple").limit(26)
        assert result.count() == 26
        objects = [a for a in result]
        assert len(objects) == 26

    def test_offset(self):
        result = PtabProceeding.objects.filter(party_name="Apple").limit(3)
        objects = [a for a in result]
        assert result.first() == objects[0]
        assert result.offset(1).first() == objects[1]
        assert result.offset(1).offset(1).first() == objects[2]


class TestPtabDocument:
    def test_filter_by_proceeding(self):
        result = PtabDocument.objects.filter(proceeding_number="IPR2016-00831")
        assert result.count() == 77

    def test_sort_by_document_number(self):
        result = PtabDocument.objects.filter(proceeding_number="IPR2016-00831").limit(3).count()
        assert result == 3


class TestPtabDecision:
    def test_get_by_proceeding(self):
        result = PtabDecision.objects.get(proceeding_number="IPR2016-00831")
        assert result.identifier == "a44c5f1557b7b60d00e66604d3668ce442d53f964aa597011cc476b4"



import pytest

from .api import PublicSearchApi


@pytest.mark.asyncio
async def test_simple_search():
    api = PublicSearchApi()
    results = await api.run_query(
        query='("6013599").pn.',
    )
    assert results.num_found == 1
    assert results.docs[0].publication_number == "6013599"



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *    Source File: patent_client/_async/epo/ops/published/model/__init__.py     *
# ********************************************************************************

from .biblio import BiblioResult, InpadocBiblio
from .fulltext import Claims, Description
from .images import ImageDocument, Images
from .search import Inpadoc, Search

__all__ = [
    "BiblioResult",
    "InpadocBiblio",
    "Claims",
    "Description",
    "Images",
    "ImageDocument",
    "Inpadoc",
    "Search",
]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *     Source File: patent_client/_async/uspto/global_dossier/model_test.py     *
# ********************************************************************************

import pytest

from .model import Document, DocumentList, GlobalDossier, GlobalDossierApplication


class TestGlobalDossier:
    def test_us_lookup(self):
        result = GlobalDossier.objects.get("16123456")
        assert result.id == "16123456"

    def test_us_lookup_docs(self):
        dossier = GlobalDossier.objects.get("16123456")
        result = dossier.applications[0].document_list
        assert len(result.office_action_docs) >= 2

    @pytest.mark.skip("issue in pytest-recording")
    def test_us_download_docs(self, tmpdir):
        dossier = GlobalDossier.objects.get("16123456")
        result = dossier.applications[0].documents[0].download(tmpdir)
        assert result.exists()

    def test_us_app_lookup(self):
        result = GlobalDossierApplication.objects.get("16123456")
        assert result.app_num == "16123456"
        assert isinstance(result, GlobalDossierApplication)

    def test_us_app_lookup_docs(self):
        result = GlobalDossierApplication.objects.get("16123456")
        document_list = result.document_list
        documents = result.documents
        office_actions = result.office_actions
        assert isinstance(document_list, DocumentList)
        assert isinstance(documents, list)
        assert isinstance(documents[0], Document)
        assert isinstance(office_actions, list)
        assert isinstance(office_actions[0], Document)

    def test_links(self):
        app = GlobalDossierApplication.objects.get("16123456")
        us_application = app.us_application
        assert (
            us_application.patent_title
            == "LEARNING ASSISTANCE DEVICE, METHOD OF OPERATING LEARNING ASSISTANCE DEVICE, LEARNING ASSISTANCE PROGRAM, LEARNING ASSISTANCE SYSTEM, AND TERMINAL DEVICE"
        )
        # Broken links to Public Patent Search
        # assert (
        #    GlobalDossierApplication.objects.get("16123456").us_publication.patent_title
        #    == "Learning assistance device, method of operating learning assistance device, learning assistance program, learning assistance system, and terminal device"
        # )
        # assert (
        #    GlobalDossierApplication.objects.get("16123456").us_patent.patent_title
        #    == "Learning assistance device, method of operating learning assistance device, learning assistance program, learning assistance system, and terminal device"
        # )
        assignments = app.us_assignments
        assert (assignments.first()).id == "46816-108"
        with pytest.raises(ValueError):
            (GlobalDossierApplication.objects.get(publication="EP1000000")).us_application
        # Broken links to Public Patent Search
        # with pytest.raises(ValueError):
        #    GlobalDossierApplication.objects.get(publication="EP1000000").us_publication
        # with pytest.raises(ValueError):
        #    GlobalDossierApplication.objects.get(publication="EP1000000").us_patent
        with pytest.raises(ValueError):
            (GlobalDossierApplication.objects.get(publication="EP1000000")).us_assignments

    def test_issue_99(self):
        app = GlobalDossierApplication.objects.get("16740760", type="application", office="US")
        assert app.app_num == "16740760"
        app = GlobalDossierApplication.objects.get("17193105")
        assert app.app_num == "17193105"



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *           Source File: patent_client/_async/uspto/ptab/api_test.py           *
# ********************************************************************************

from .api import PtabApi


class TestPtabApi:
    def test_can_get_document(self):
        page = PtabApi.get_documents(document_identifier="170680401")
        assert len(page.docs) == 1

    def test_can_get_document_range(self):
        page = PtabApi.get_documents(
            proceeding_number="IPR2022-00037",
            document_filing_date=("2023-01-01", "2023-09-15"),
        )
        assert len(page.docs) == 25

    def test_can_get_proceeding(self):
        page = PtabApi.get_proceedings(proceeding_number="IPR2022-00037")
        assert len(page.docs) == 1



import asyncio
import json
from copy import deepcopy
from pathlib import Path

import httpx

from patent_client._async.http_client import PatentClientSession

from .model import PublicSearchBiblioPage, PublicSearchDocument


class UsptoException(Exception):
    pass


def force_list(obj):
    if not isinstance(obj, list):
        return [
            obj,
        ]
    return obj


class PublicSearchApi:
    def __init__(self):
        self.client = PatentClientSession(
            headers={
                "X-Requested-With": "XMLHttpRequest",
                "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
                "Origin": "https://ppubs.uspto.gov",
                "Referer": "https://ppubs.uspto.gov/pubwebapp/",
                "Pragma": "no-cache",
                "Cache-Control": "no-cache",
                "Priority": "u=1, i",
            },
            http2=True,
            follow_redirects=True,
        )
        self.session = dict()
        self.case_id = None
        self.queries = dict()
        self.search_query = json.loads((Path(__file__).parent / "search_query.json").read_text())

    async def run_query(
        self,
        query,
        start=0,
        limit=500,
        sort="date_publ desc",
        default_operator="OR",
        sources=["US-PGPUB", "USPAT", "USOCR"],
        expand_plurals=True,
        british_equivalents=True,
    ) -> "PublicSearchBiblioPage":
        if self.case_id is None:
            await self.get_session()

        data = deepcopy(self.search_query)
        data["start"] = start
        data["pageCount"] = limit
        data["sort"] = sort
        data["query"]["caseId"] = self.case_id
        data["query"]["op"] = default_operator
        data["query"]["q"] = query
        data["query"]["queryName"] = query
        data["query"]["userEnteredQuery"] = query
        data["query"]["databaseFilters"] = [
            {"databaseName": s, "countryCodes": []} for s in sources
        ]
        data["query"]["plurals"] = expand_plurals
        data["query"]["britishEquivalents"] = british_equivalents

        counts = await self.make_request(
            "POST",
            "https://ppubs.uspto.gov/dirsearch-public/searches/counts",
            json=data["query"],
        )
        counts.raise_for_status()
        search_url = "https://ppubs.uspto.gov/dirsearch-public/searches/searchWithBeFamily"
        query_response = await self.make_request("POST", search_url, json=data)
        query_response.raise_for_status()
        result = query_response.json()
        if result.get("error", None) is not None:
            raise UsptoException(
                f"Error #{result['error']['errorCode']}\n{result['error']['errorMessage']}"
            )
        return PublicSearchBiblioPage.model_validate(result)

    async def make_request(self, method, url, **kwargs):
        response = await self.client.request(method, url, **kwargs)
        if response.status_code == 403:
            await self.get_session()
            response = await self.client.request(method, url, **kwargs)
        if response.status_code == 429:
            wait_time = int(response.headers["x-rate-limit-retry-after-seconds"]) + 1
            await asyncio.sleep(wait_time)
            response = await self.client.request(method, url, **kwargs)
        return response

    async def get_document(self, bib) -> "PublicSearchDocument":
        url = f"https://ppubs.uspto.gov/dirsearch-public/internal/patents/{bib.guid}/highlight"
        params = {
            "queryId": 1,
            "source": bib.type,
            "includeSections": True,
            "uniqueId": None,
        }
        response = await self.make_request("GET", url, params=params)
        response.raise_for_status()
        return PublicSearchDocument.model_validate(response.json())

    async def get_session(self):
        self.client.cookies = httpx.Cookies()
        response = await self.client.get("https://ppubs.uspto.gov/pubwebapp/")
        url = "https://ppubs.uspto.gov/dirsearch-public/users/me/session"
        response = await self.client.post(
            url,
            json=-1,
            headers={
                "X-Access-Token": "null",
                "referer": "https://ppubs.uspto.gov/pubwebapp/",
            },
        )  # json=str(random.randint(10000, 99999)))
        self.session = response.json()
        self.case_id = self.session["userCase"]["caseId"]
        self.access_token = response.headers["X-Access-Token"]
        self.client.headers["X-Access-Token"] = self.access_token
        return self.session

    async def _request_save(self, obj):
        page_keys = [
            f"{obj.image_location}/{i:0>8}.tif"
            for i in range(1, obj.document_structure.page_count + 1)
        ]
        response = await self.client.post(
            "https://ppubs.uspto.gov/dirsearch-public/internal/print/imageviewer",
            json={
                "caseId": self.case_id,
                "pageKeys": page_keys,
                "patentGuid": obj.guid,
                "saveOrPrint": "save",
                "source": obj.type,
            },
        )
        if response.status_code == 500:
            raise UsptoException(response.text)
        return response.text

    async def download_image(self, obj, path="."):
        out_path = Path(path).expanduser() / f"{obj.guid}.pdf"
        if out_path.exists():
            return out_path
        if self.case_id is None:
            await self.get_session()
        try:
            print_job_id = await self._request_save(obj)
        except httpx.HTTPStatusError:
            await self.get_session()
            print_job_id = await self._request_save(obj)
        while True:
            response = await self.client.post(
                "https://ppubs.uspto.gov/dirsearch-public/internal/print/print-process",
                json=[
                    print_job_id,
                ],
            )
            response.raise_for_status()
            print_data = response.json()
            if print_data[0]["printStatus"] == "COMPLETED":
                break
            await asyncio.sleep(1)
        pdf_name = print_data[0]["pdfName"]
        with out_path.open("wb") as f:
            try:
                request = self.client.build_request(
                    "GET",
                    f"https://ppubs.uspto.gov/dirsearch-public/internal/print/save/{pdf_name}",
                )
                response = await self.client.send(request, stream=True)
                response.raise_for_status()
                async for chunk in response.aiter_bytes():
                    if chunk:
                        f.write(chunk)
            except httpx.HTTPStatusError as e:
                await response.aclose()
                raise e
        return out_path



import pytest

from .model import USApplication, USApplicationBiblio


@pytest.mark.asyncio
async def test_all_apps():
    assert await USApplication.objects.filter(query=dict()).count() > 1000


@pytest.mark.asyncio
async def test_get_one_app():
    app = await USApplication.objects.get(q="applicationNumberText:16123456")
    assert app is not None
    assert app.appl_id == "16123456"


@pytest.mark.asyncio
async def test_get_app_from_search_result():
    application = await USApplication.objects.get(q="applicationNumberText:16123456")
    assert application.appl_id == "16123456"


@pytest.mark.asyncio
async def test_get_app_biblio_from_search_result():
    result = await USApplicationBiblio.objects.get(q="applicationNumberText:16123456")
    biblio = await result.biblio
    assert biblio.appl_id == "16123456"


@pytest.mark.asyncio
async def test_get_continuity_from_search_result():
    result = await USApplicationBiblio.objects.get(q="applicationNumberText:16123456")
    continuity = await result.continuity
    assert len(continuity.child_continuity) > 0


@pytest.mark.asyncio
async def test_get_documents_from_search_result():
    result = await USApplicationBiblio.objects.get(q="applicationNumberText:16123456")
    documents = await result.documents
    assert await documents.count() > 0


@pytest.mark.asyncio
async def test_simple_keyword_searches():
    result = await USApplication.objects.get("16123456")
    assert result.appl_id == "16123456"


@pytest.mark.asyncio
async def test_combination_search():
    result = USApplication.objects.filter(
        invention_title="Hair Dryer", filing_date_gte="2020-01-01"
    )
    assert await result.count() > 5


@pytest.mark.asyncio
async def test_can_get_old_applications():
    result = await USApplication.objects.get("14230558")
    assert result.appl_id == "14230558"
    result = await USApplicationBiblio.objects.get("14230558")
    assert result.appl_id == "14230558"


@pytest.mark.asyncio
async def test_can_get_pct_application():
    result = await USApplication.objects.get("PCT/US07/19317")
    assert result.appl_id == "PCT/US07/19317"


@pytest.mark.asyncio
async def test_can_get_by_customer_number():
    result = USApplication.objects.filter(customer_number="31625")
    assert await result.count() > 0



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *         Source File: patent_client/_async/epo/ops/legal/api_test.py          *
# ********************************************************************************

from pathlib import Path

from .api import LegalApi

fixtures = Path(__file__).parent / "fixtures"


def test_async_example():
    result = LegalApi.get_legal("EP1000000A1")
    assert str(result.publication_reference) == "EP1000000A1"
    assert len(result.events) >= 50



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *           Source File: patent_client/_async/uspto/odp/api_test.py            *
# ********************************************************************************

import pytest

from .api import ODPApi
from .model import SearchRequest


@pytest.fixture
def odp_api():
    return ODPApi()


def test_post_search(odp_api):
    search_request = SearchRequest(q="firstNamedApplicant:STMicroelectronics S.A.")
    response = odp_api.post_search(search_request)
    assert response["count"] > 0, "Expected at least one result"


def test_get_search(odp_api):
    search_request = SearchRequest(q="firstNamedApplicant:STMicroelectronics S.A.")
    response = odp_api.get_search(search_request)
    assert response["count"] > 0, "Expected at least one result"


def test_get_application_data(odp_api):
    application_id = "15123456"
    application = odp_api.get_application_data(application_id)
    assert application.appl_id is not None, "Expected patent data"


def test_get_application_basic_data(odp_api):
    application_id = "15123456"
    application = odp_api.get_application_biblio_data(application_id)
    assert application.appl_id is not None, "Expected basic patent data"


def test_get_patent_term_adjustment_data(odp_api):
    application_id = "16123456"
    response = odp_api.get_patent_term_adjustment_data(application_id)
    assert response.adjustment_total_quantity is not None, "Expected patent term adjustment data"


def test_get_assignments(odp_api):
    application_id = "15123456"
    response = odp_api.get_assignments(application_id)
    assert len(response) > 0, "Expected at least one assignment"


def test_get_attorney_data(odp_api):
    application_id = "15123456"
    response = odp_api.get_attorney_data(application_id)
    assert response.attorneys is not None, "Expected attorney data"


def test_get_continuity_data(odp_api):
    application_id = "15123456"
    response = odp_api.get_continuity_data(application_id)
    assert response.parent_continuity is not None, "Expected continuity data"


def test_get_foreign_priority_data(odp_api):
    application_id = "16123456"
    response = odp_api.get_foreign_priority_data(application_id)
    assert len(response) > 0, "Expected at least one foreign priority"


def test_get_transactions(odp_api):
    application_id = "15123456"
    response = odp_api.get_transactions(application_id)
    assert len(response) > 0, "Expected at least one transaction"


def test_get_documents(odp_api):
    application_id = "15123456"
    response = odp_api.get_documents(application_id)
    assert len(response) > 0, "Expected at least one document"



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *             Source File: patent_client/_async/epo/ops/session.py             *
# ********************************************************************************

import base64
import datetime as dt
import logging
from typing import Sequence

import hishel
import httpx
from httpcore import Request, Response

from patent_client import SETTINGS
from patent_client._sync.http_client import PatentClientSession
from patent_client.session import CACHE_DIR

logger = logging.getLogger(__name__)
NS = {
    "http://ops.epo.org": None,
    "http://www.epo.org/exchange": None,
    "http://www.epo.org/fulltext": None,
    "http://www.epo.org/register": None,
}


class OpsAuthenticationError(Exception):
    pass


class OpsForbiddenError(Exception):
    pass


class OpsFairUseError(Exception):
    pass


class OpsController(hishel.Controller):
    def __init__(
        self,
        cacheable_methods: Sequence[str] = ("GET", "HEAD"),
        cacheable_status_codes: Sequence[int] = (200,),
    ):
        super().__init__()
        self.cacheable_methods = cacheable_methods
        self.cacheable_status_codes = cacheable_status_codes

    def is_cachable(self, request: Request, response: Response) -> bool:
        return (
            request.method in self.cacheable_methods
            and response.status_code in self.cacheable_status_codes
        )

    def construct_response_from_cache(self, request, response, original_request):
        return response


class OpsAuth(httpx.Auth):
    requires_response_body = True
    auth_url = "https://ops.epo.org/3.2/auth/accesstoken"

    def __init__(self, key: str, secret: str):
        self.key = key
        self.secret = secret
        self.authorization_header = "<unset>"

    def auth_flow(self, request):
        request.headers["Authorization"] = self.authorization_header
        response = yield request
        if response.status_code == 400:
            response = yield self.build_refresh_request()
            if response.status_code != 200:
                logger.debug(f"EPO Authentication Error!\n{response.text}")
                raise OpsAuthenticationError(
                    "Failed to authenticate with EPO OPS! Please check your credentials. See the setup instructions at https://patent-client.readthedocs.io/en/stable/getting_started.html"
                )
            data = response.json()
            self.expires = dt.datetime.fromtimestamp(int(data["issued_at"]) / 1000) + dt.timedelta(
                seconds=int(data["expires_in"])
            )
            self.authorization_header = f"Bearer {data['access_token']}"
            request.headers["Authorization"] = self.authorization_header
            yield request

    def build_refresh_request(self):
        token = base64.b64encode(f"{self.key}:{self.secret}".encode())
        return httpx.Request(
            "post",
            self.auth_url,
            headers={"Authorization": token},
            data={"grant_type": "client_credentials"},
        )


ops_transport = hishel.CacheTransport(
    transport=httpx.HTTPTransport(
        verify=False,
        http2=True,
        retries=3,
    ),
    storage=hishel.FileStorage(base_path=CACHE_DIR, ttl=60 * 60 * 24 * 3),
    controller=OpsController(),
)


def handle_response(response):
    if response.status_code == 403:
        raise OpsForbiddenError("Forbidden")
    return response


session = PatentClientSession(
    transport=ops_transport,
    auth=OpsAuth(key=SETTINGS.epo_api_key, secret=SETTINGS.epo_api_secret),
    event_hooks={
        "response": [
            handle_response,
        ]
    },
)



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *          Source File: patent_client/_async/epo/ops/published/api.py          *
# ********************************************************************************

import logging
from io import BytesIO

from yankee.data import AttrDict

from ..session import session
from .model.biblio import BiblioResult
from .model.fulltext import Claims, Description
from .model.images import Images
from .model.search import Search

logger = logging.getLogger(__name__)


class PublishedBiblioApi:
    @classmethod
    def get_constituents(
        cls, number, doc_type="publication", format="docdb", constituents=("biblio",)
    ):
        """Published Data Constituents API
        number: document number to search
        doc_type: document type (application / publication)
        format: document number format (original / docdb / epodoc)
        constituents: what data to retrieve. Can be combined. (biblio / abstract / full-cycle)
        """
        base_url = (
            f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/"
        )
        if isinstance(constituents, str):
            constituents = (constituents,)
        url = base_url + ",".join(constituents)
        response = session.get(url)
        return response.text

    @classmethod
    def get_biblio(cls, number, doc_type="publication", format="docdb") -> "BiblioResult":
        text = cls.get_constituents(number, doc_type, format, constituents="biblio")
        return BiblioResult.model_validate(text)

    @classmethod
    def get_abstract(cls, number, doc_type="publication", format="docdb"):
        return cls.get_constituents(number, doc_type, format, constituents="abstract")

    @classmethod
    def get_full_cycle(cls, number, doc_type="publication", format="docdb"):
        return cls.get_constituents(number, doc_type, format, constituents="full-cycle")


class PublishedFulltextApi:
    fulltext_jurisdictions = "EP, WO, AT, BE, BG, CA, CH, CY, CZ, DK, EE, ES, FR, GB, GR, HR, IE, IT, LT, LU, MC, MD, ME, NO, PL, PT, RO, RS, SE, SK".split(
        ", "
    )

    @classmethod
    def get_fulltext_result(
        cls, number, doc_type="publication", format="docdb", inquiry="fulltext"
    ):
        """Published Fulltext API
        number: document number to search
        doc_type: document type (application / publication)
        format: document number format (original / docdb / epodoc)
        inquiry: what data to retrieve. Can be combined. (fulltext / description / claims)
        """
        url = f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/{inquiry}"
        if number[:2] not in cls.fulltext_jurisdictions:
            raise ValueError(
                f"Fulltext Is Not Available For Country Code {number[:2]}. Fulltext is only available in {', '.join(cls.fulltext_jurisdictions)}"
            )
        response = session.get(url)
        response.raise_for_status
        return response.text

    @classmethod
    def get_description(cls, number, doc_type="publication", format="docdb"):
        text = cls.get_fulltext_result(
            number, doc_type="publication", format="docdb", inquiry="description"
        )
        return Description.model_validate(text)

    @classmethod
    def get_claims(cls, number, doc_type="publication", format="docdb") -> "Claims":
        text = cls.get_fulltext_result(
            number, doc_type="publication", format="docdb", inquiry="claims"
        )
        return Claims.model_validate(text)


class PublishedSearchApi:
    @classmethod
    def search(cls, query, start=1, end=100):
        base_url = "http://ops.epo.org/3.2/rest-services/published-data/search"
        range = f"{start}-{end}"
        logger.debug(f"OPS Search Endpoint - Query: {query}\nRange: {start}-{end}")
        response = session.get(base_url, params={"Range": range, "q": query})
        if response.status_code == 404:
            return AttrDict.convert(
                {
                    "query": "query",
                    "num_results": 0,
                    "begin": start,
                    "end": end,
                    "results": list(),
                }
            )
        return Search.model_validate(response.text)


class PublishedImagesApi:
    @classmethod
    def get_images(cls, number, doc_type="publication", format="docdb"):
        base_url = f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/images"
        response = session.get(base_url)
        return Images.model_validate(response.text)

    @classmethod
    def get_page_image(cls, country, number, kind, image_type, page_number, image_format="pdf"):
        response = session.get(
            f"https://ops.epo.org/3.2/rest-services/published-data/images/{country}/{number}/{kind}/{image_type}.{image_format}",
            params={"Range": page_number},
            extensions={"force_cache": True},
        )
        return BytesIO(response.content)

    @classmethod
    def get_page_image_from_link(cls, link, page_number, image_format="pdf"):
        response = session.get(
            f"https://ops.epo.org/3.2/rest-services/{link}.{image_format}",
            params={"Range": page_number},
            extensions={"force_cache": True},
        )
        return BytesIO(response.content)


class PublishedApi:
    biblio = PublishedBiblioApi
    fulltext = PublishedFulltextApi
    search = PublishedSearchApi
    images = PublishedImagesApi



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *         Source File: patent_client/_async/uspto/assignment/model.py          *
# ********************************************************************************

import datetime
import typing as tp
import warnings
from pathlib import Path

from dateutil.parser import isoparse
from pydantic import BeforeValidator, ConfigDict, Field, field_validator
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated

from patent_client.util.pydantic_util import BaseModel

if tp.TYPE_CHECKING:
    from ..peds.model import USApplication
    from ..public_search.model import Patent, PublishedApplication
UsptoDate = Annotated[
    datetime.date,
    BeforeValidator(lambda x: datetime.datetime.strptime(x, "%Y%m%d").date()),
]
YNBool = Annotated[bool, BeforeValidator(lambda x: x == "Y")]


def parse_datetime(string):
    dt = isoparse(string)
    if dt.year == 1 and dt.month == 1 and dt.day == 1:
        return None
    return dt


def parse_date(string):
    dt = parse_datetime(string)
    if dt is None:
        return None
    return dt.date()


DatetimeAsDate = Annotated[tp.Optional[datetime.date], BeforeValidator(lambda x: parse_date(x))]


class AbstractAssignmentModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        str_strip_whitespace=True,
    )


class Assignor(AbstractAssignmentModel):
    name: str = Field(alias="patAssignorName")
    execution_date: DatetimeAsDate = Field(alias="patAssignorExDate")
    acknowledgement_date: DatetimeAsDate = Field(alias="patAssignorDateAck")


class Assignee(AbstractAssignmentModel):
    name: str = Field(alias="patAssigneeName")
    address: str = Field(alias="patAssigneeAddress")


class Property(AbstractAssignmentModel):
    invention_title: str
    invention_title_lang: str
    appl_num: str
    filing_date: tp.Optional[DatetimeAsDate] = None
    intl_publ_date: tp.Optional[DatetimeAsDate] = None
    intl_reg_num: tp.Optional[str] = None
    inventors: tp.Optional[str] = None
    issue_date: tp.Optional[DatetimeAsDate] = None
    pat_num: tp.Optional[str] = None
    pct_num: tp.Optional[str] = None
    publ_date: tp.Optional[DatetimeAsDate] = None
    publ_num: tp.Optional[str] = None

    @property
    def application(self) -> tp.Optional["USApplication"]:
        """The related US Application"""
        try:
            appl_id = getattr(self, "appl_id", None)
            if appl_id is not None:
                return self._get_model("patent_client.uspto.peds.model.USApplication").objects.get(
                    appl_id=appl_id
                )
            appl_id = getattr(self, "pct_num", None)
            if appl_id is not None:
                return self._get_model("patent_client.uspto.peds.model.USApplication").objects.get(
                    appl_id=appl_id
                )
            pub_num = getattr(self, "publ_num", None)
            if pub_num is not None:
                return self._get_model("patent_client.uspto.peds.model.USApplication").objects.get(
                    app_early_pub_number=pub_num
                )
            pat_num = getattr(self, "pat_num", None)
            if pat_num is not None:
                return self._get_model("patent_client.uspto.peds.model.USApplication").objects.get(
                    patent_number=pat_num
                )
        except Exception:
            pass
        warnings.warn(f"Unable to find application for {self}")
        return None

    @property
    def patent(self) -> "Patent":
        """The related US Patent, if any"""
        return self._get_model("patent_client.uspto.public_search.model.Patent").objects.get(
            publication_number=self.pat_num
        )

    @property
    def publication(
        self,
    ) -> "PublishedApplication":
        """The related US Publication, if any"""
        return self._get_model(
            "patent_client.uspto.peds.fulltext.patent.model.PublishedApplication"
        ).objects.get(publication_number=self.publ_num)


class Correspondent(AbstractAssignmentModel):
    name: str
    address: str


class Assignment(AbstractAssignmentModel["AssignmentManager"]):
    id: str
    date_produced: tp.Optional[UsptoDate] = None
    action_key_code: tp.Optional[str] = None
    transaction_date: tp.Optional[DatetimeAsDate] = None
    last_update_date: DatetimeAsDate
    purge_indicator: YNBool
    recorded_date: DatetimeAsDate
    page_count: int
    conveyance_text: str
    assignment_record_has_images: YNBool
    attorney_dock_num: tp.Optional[str] = None
    pat_assignor_earliest_ex_date: DatetimeAsDate
    correspondent: Correspondent
    assignors: tp.List[Assignor]
    assignees: tp.List[Assignee]
    properties: tp.List[Property]

    @field_validator("conveyance_text")
    @classmethod
    def remove_see_document(cls, v) -> str:
        return v.replace(" (SEE DOCUMENT FOR DETAILS).", "").strip()

    @property
    def reel_frame(self):
        return self.id.split("-")

    @property
    def image_url(self):
        reel, frame = self.reel_frame
        reel = reel.rjust(6, "0")
        frame = frame.rjust(4, "0")
        return f"http://legacy-assignments.uspto.gov/assignments/assignment-pat-{reel}-{frame}.pdf"

    def download(self, path: tp.Optional[tp.Union[str, Path]] = None):
        """asynchronously downloads the PDF associated with the assignment to the current working directory"""
        from .api import client

        return client.download(self.image_url, path=path)


class AssignmentPage(AbstractAssignmentModel):
    num_found: int
    docs: list[Assignment]

    def __len__(self):
        return self.num_found



import functools


class AsyncProxyObject:
    """
    Proxy for an async object that ensures the object is awaited and stored.
    """

    def __init__(self, coroutine, attr=None, index=None):
        self._coroutine = coroutine
        self._object = None
        self._attr = attr
        self._index = index

    async def _ensure_object(self):
        if self._object is None:
            self._object = await self._coroutine
        if self._index is not None:
            return self._object[self._index]
        elif self._attr is not None:
            return getattr(self._object, self._attr)
        return self._object

    def __getattr__(self, item):
        return AsyncProxyObject(self._ensure_object(), attr=item)

    def __getitem__(self, item):
        return AsyncProxyObject(self._ensure_object(), index=item)

    def __await__(self):
        return self._ensure_object().__await__()


def async_proxy(_func=None, *, attr=None):
    def decorator_async_proxy(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return AsyncProxyObject(func(*args, **kwargs), attr=attr)

        return wrapper

    if _func is None:
        return decorator_async_proxy
    else:
        return decorator_async_proxy(_func)



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *                   Source File: patent_client/_async/odp.py                   *
# ********************************************************************************

# A simple separate namespace for the USPTO's Open Data Portal.
from patent_client._sync.uspto.odp.model import USApplication, USApplicationBiblio

__all__ = ["USApplication", "USApplicationBiblio"]



from __future__ import annotations

from collections import OrderedDict
from copy import deepcopy
from enum import Enum
from itertools import chain
from typing import TYPE_CHECKING, AsyncIterator, Generic, Iterator, TypeVar, Union

from typing_extensions import Self
from yankee.data import Collection

if TYPE_CHECKING:
    pass

ModelType = TypeVar("ModelType")


class OrderDirection(str, Enum):
    ASC = "asc"
    DESC = "desc"


class ManagerConfig:
    """
    Manager Configuration Class

    This class is designed to store and manage configuration settings for a manager object. It allows for the customization of query parameters and options to tailor data retrieval processes. The attributes of this class include:

    Attributes:
        filter (OrderedDict[str, list]): An ordered dictionary to store filter conditions for queries. The keys represent the field names, and the values represent the filter criteria.
        order_by (list[tuple[str, OrderDirection]]): A list of tuples specifying the ordering of query results. Each tuple contains a field name and the direction ('asc' for ascending, 'desc' for descending) of the sort.
        options (dict[str, str]): A dictionary to store additional options that may affect the query or its results.
        limit (int | None): An optional integer specifying the maximum number of results to return. If None, no limit is applied.
        offset (int): An integer specifying the offset from the start of the result set. Used for pagination.
        annotations (list[tuple[str, str]]): A list of tuples for annotating the results with extra information. Each tuple contains a field name and an annotation.

    The class also includes a method for comparing two `ManagerConfig` instances for equality based on their attributes.
    """

    def __init__(self):
        self.filter: OrderedDict[str, list] = OrderedDict()
        self.order_by: list[tuple[str, OrderDirection]] = list()
        self.options: dict[str, str] = dict()
        self.limit: int | None = None
        self.offset: int = 0
        self.annotations: list[tuple[str, str]] = list()

    def __eq__(self, other):
        return (
            self.filter == other.filter
            and self.order_by == other.order_by
            and self.options == other.options
            and self.limit == other.limit
            and self.offset == other.offset
            and self.annotations == other.annotations
        )


class BaseManager(Collection, Generic[ModelType]):
    default_filter: str = ""

    def __init__(self, config=None):
        self.config = config or ManagerConfig()

    def __eq__(self, other) -> bool:
        return bool(self.config == other.config and isinstance(self, type(other)))

    def __add__(self, other):
        return Collection(chain(self, other))

    # Manager Modification Functions

    def filter(self, *args, **kwargs) -> Self:
        """Apply a new filtering condition"""
        mger = deepcopy(self)
        if args:
            kwargs[self.default_filter] = args

        for key, value in kwargs.items():
            if isinstance(value, (str, dict, int, float)):
                kwargs[key] = [value]
            else:
                kwargs[key] = list(value)

        for key, value in kwargs.items():
            if key in mger.config.filter:
                mger.config.filter[key].extend(value)
            else:
                mger.config.filter[key] = value

        return mger

    def order_by(self, *args) -> Self:
        """Specify the order that argument should be returned in"""
        mger = deepcopy(self)
        mger.config.order_by = args
        return mger

    def option(self, **kwargs) -> Self:
        """Set a key:value option on the manager"""
        mger = deepcopy(self)
        mger.config.options = {**mger.config.options, **kwargs}
        return mger

    def limit(self, limit) -> Self:
        """Limit the number of records that are returned"""
        mger = deepcopy(self)
        mger.config.limit = limit
        return mger

    def offset(self, offset) -> Self:
        """Specify the number of records from the beginning from which to apply an offset"""
        mger = deepcopy(self)
        mger.config.offset = self.config.offset + offset
        return mger

    # Basic Manager Fetching

    def count(self) -> int:
        """Returns number of records in the QuerySet. Alias for len(self)"""
        return len(self)

    def all(self) -> Manager[ModelType]:
        """Return self. Does nothing"""
        return self


class Manager(BaseManager, Generic[ModelType]):
    """
    Manager Class (Synchronous)

    This class is essentially a configurable generator. It is intended to be initialized
    as an empty object at Model.objects. Users can then call methods to modify the manager.
    All methods should return a brand-new manager with the appropriate parameters re-set.
    The manager's attributes are stored in a dictionary at Manager.config.

    """

    default_filter: str = ""

    def __init__(self, config=None):
        self.config = config or ManagerConfig()

    # Manager Iteration / Slicing

    def __iter__(self) -> Iterator[ModelType]:
        return self._get_results()

    def __getitem__(self, key: Union[slice, int]) -> Union[Manager[ModelType], ModelType]:
        if isinstance(key, slice):
            if key.step is not None:
                raise AttributeError("Step is not supported")
            start = key.start if key.start else 0
            start = len(self) + start if start < 0 else start
            stop = key.stop if key.stop else len(self)
            stop = len(self) + stop if stop < 0 else stop
            mger = self.offset(start + self.config.offset)
            mger = mger.limit(stop - start)
            return mger
        return self.offset(key).first()

    # Basic Manager Attributes

    def __len__(self) -> int:
        return self.count()

    def __eq__(self, other) -> bool:
        return bool(self.config == other.config and isinstance(self, type(other)))

    def __add__(self, other):
        return Collection(chain(self, other))

    def first(self) -> ModelType:
        """Get the first object in the manager"""
        return next(self.limit(1).__iter__())

    def get(self, *args, **kwargs) -> ModelType:
        """If the critera results in a single record, return it, else raise an exception"""
        mger = self.filter(*args, **kwargs)
        length = len(mger)
        if length > 1:
            raise ValueError("More than one document found!")
        if length == 0:
            raise ValueError("No documents found!")
        return mger.first()


class AsyncManager(BaseManager, Generic[ModelType]):
    """
    Manager Class (Asynchronous)

    This class is essentially a configurable generator. It is intended to be initialized
    as an empty object at Model.objects. Users can then call methods to modify the manager.
    All methods should return a brand-new manager with the appropriate parameters re-set.
    The manager's attributes are stored in a dictionary at Manager.config.

    """

    default_filter: str = ""

    def __init__(self, config=None):
        self.config = config or ManagerConfig()

    # Manager Iteration / Slicing

    def __aiter__(self) -> AsyncIterator[ModelType]:
        return self._get_results()

    async def _get_results(self) -> AsyncIterator[ModelType]:
        raise NotImplementedError(
            f"This method must be defined in a subclass of {self.__class__.__name__}"
        )

    async def __getitem__(self, key: Union[slice, int]) -> Union[Manager[ModelType], ModelType]:
        if isinstance(key, slice):
            if key.step is not None:
                raise AttributeError("Step is not supported")
            count = await self.count()
            start = key.start if key.start else 0
            start = count + start if start < 0 else start
            stop = key.stop if key.stop else count
            stop = count + stop if stop < 0 else stop
            mger = self.offset(start + self.config.offset)
            mger = mger.limit(stop - start)
            return mger
        return await self.offset(key).first()

    # Basic Manager Attributes

    def __len__(self) -> int:
        raise NotImplementedError(
            f"This method is not implemented for the AsyncManager class. Use await {self.__class__.__name__}.count() instead."
        )

    async def count(self) -> int:
        """Returns number of records in the QuerySet. Alias for len(self)"""
        return NotImplemented(
            f"This method must be defined in a subclass of {self.__class__.__name__}"
        )

    async def len(self) -> int:
        """Returns number of records in the QuerySet. Alias for self.count()"""
        return await self.count()

    async def first(self) -> ModelType:
        """Get the first object in the manager"""
        try:
            return await anext(self.limit(1).__aiter__())
        except NameError:  # anext was added in 3.10
            async for doc in self.limit(1):
                return doc

    async def get(self, *args, **kwargs) -> ModelType:
        """If the critera results in a single record, return it, else raise an exception"""
        mger = self.filter(*args, **kwargs)
        length = await mger.count()
        if length > 1:
            raise ValueError("More than one document found!")
        if length == 0:
            raise ValueError("No documents found!")
        return await mger.first()

    async def to_list(self) -> list[ModelType]:
        """Return a list of all objects in the manager"""
        return [item async for item in self]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *            Source File: patent_client/_async/epo/ops/legal/api.py            *
# ********************************************************************************

from ..session import session
from .model import Legal


class LegalApi:
    @classmethod
    def get_legal(cls, doc_number, doc_type="publication", format="docdb"):
        url = f"http://ops.epo.org/3.2/rest-services/legal/{doc_type}/{format}/{doc_number}"
        response = session.get(url)
        return Legal.model_validate(response.text)



from . import odp
from .epo.ops.published.model import Inpadoc
from .uspto.assignment.model import Assignment
from .uspto.global_dossier.model import GlobalDossier, GlobalDossierApplication
from .uspto.peds.model import USApplication
from .uspto.ptab.model import PtabDecision, PtabDocument, PtabProceeding
from .uspto.public_search.model import (
    Patent,
    PatentBiblio,
    PublicSearchBiblio,
    PublicSearchDocument,
    PublishedApplication,
    PublishedApplicationBiblio,
)

__all__ = [
    "Inpadoc",
    "Assignment",
    "USApplication",
    "PtabDecision",
    "PtabDocument",
    "PtabProceeding",
    "GlobalDossier",
    "GlobalDossierApplication",
    "PublicSearchBiblio",
    "PublicSearchDocument",
    "Patent",
    "PatentBiblio",
    "PublishedApplication",
    "PublishedApplicationBiblio",
    "odp",
]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *       Source File: patent_client/_async/epo/ops/number_service/api.py        *
# ********************************************************************************

from ..session import session
from .model import NumberServiceResult


class NumberServiceException(Exception):
    pass


class NumberServiceApi:
    @classmethod
    def convert_number(
        cls,
        number,
        doc_type="publication",
        input_format="original",
        output_format="docdb",
    ) -> NumberServiceResult:
        """Number Conversion Service
        number: number to convert. Must include at least country and doc number,
            can optionally include kind code or publication date. e.g.
            MD.20050130
            MD.20050130.A
            MD.20050130.A.20050130
            JP.2006-147056.A.20060526
            JP.2006147056.A.20060526
            US.(08/921,321).A.19970829
        doc_type: type of document (publication / application / priority)
        input_format: input type (original / docdb / epodoc)
        output_format: output type (original / docdb / epodoc)
        """
        response = session.get(
            f"http://ops.epo.org/3.2/rest-services/number-service/{doc_type}/{input_format}/{number}/{output_format}",
            extensions={"force_cache": True},
        )
        result = NumberServiceResult.model_validate(response.text)
        errors = [m for m in result.messages if m["kind"] == "ERROR"]
        if errors:
            raise NumberServiceException("\n".join(str(e) for e in errors))
        return result



import datetime
import json
from pathlib import Path

import pytest

from .model import (
    AssigneeAddress,
    Assignment,
    Continuity,
    CustomerNumber,
    Document,
    ForeignPriority,
    SearchRequest,
    TermAdjustment,
    TermAdjustmentHistory,
    Transaction,
    USApplication,
    USApplicationBiblio,
)


@pytest.fixture
def fixture_dir():
    return Path(__file__).parent / "fixtures"


def test_continuity(fixture_dir):
    data = json.loads((fixture_dir / "continuity.json").read_text())
    continuity = Continuity(**data)
    assert continuity.count == 1
    assert continuity.request_identifier == "50b0e4d5-98f0-4f5a-a550-5c45a9a52109"
    assert len(continuity.parent_continuity) == 2
    assert continuity.parent_continuity[0].application_status_code == 161
    assert continuity.parent_continuity[0].claim_type_code == "NST"
    assert continuity.parent_continuity[0].filing_date == datetime.date(2015, 4, 22)
    assert (
        continuity.parent_continuity[0].application_status_description
        == "Abandoned  --  Failure to Respond to an Office Action"
    )
    assert continuity.parent_continuity[0].claim_type_description == "is a National Stage Entry of"
    assert continuity.parent_continuity[0].parent_application_id == "PCT/US15/27322"
    assert continuity.parent_continuity[0].child_application_id == "15123456"
    assert len(continuity.child_continuity) == 1
    assert continuity.child_continuity[0].application_status_code == 160
    assert continuity.child_continuity[0].claim_type_code == "CON"
    assert continuity.child_continuity[0].filing_date == datetime.date(2018, 9, 13)
    assert (
        continuity.child_continuity[0].application_status_description
        == "Abandoned  --  Incomplete Application (Pre-examination)"
    )
    assert continuity.child_continuity[0].claim_type_description == "is a Continuation of"
    assert continuity.child_continuity[0].parent_application_id == "15123456"
    assert continuity.child_continuity[0].child_application_id == "16132008"


def test_document(fixture_dir):
    data = json.loads((fixture_dir / "documents.json").read_text())
    for doc_data in data["documentBag"]:
        document = Document(**doc_data)
        assert document.appl_id == "16330077"
        assert isinstance(document.mail_date, datetime.datetime)
        assert document.document_identifier is not None
        assert document.document_code is not None
        assert document.document_code_description is not None
        assert document.direction_category in ["INCOMING", "OUTGOING", "INTERNAL"]
        assert isinstance(document.download_option_bag, list)
        for download_option in document.download_option_bag:
            assert download_option["mimeTypeIdentifier"] in ["PDF", "XML", "MS_WORD"]
            assert download_option["downloadUrl"].startswith(
                "https://beta-api.uspto.gov/api/v1/download/applications/16330077/"
            )
            assert isinstance(download_option["pageTotalQuantity"], int)


def test_assignment(fixture_dir):
    data = json.loads((fixture_dir / "assignment.json").read_text())
    assignment_bag = data["patentBag"][0]["assignmentBag"]
    for assignment_data in assignment_bag:
        assignment = Assignment(**assignment_data)
        assert isinstance(assignment.assignment_received_date, datetime.date)
        assert isinstance(assignment.assignment_recorded_date, datetime.date)
        assert isinstance(assignment.assignment_mailed_date, datetime.date)
        assert assignment.frame_number is not None
        assert assignment.page_number > 0
        assert assignment.reel_number_frame_number is not None
        assert assignment.reel_number is not None
        assert assignment.conveyance_text is not None
        for assignor in assignment.assignor_bag:
            assert isinstance(assignor.execution_date, datetime.date)
            assert assignor.assignor_name is not None
        for assignee in assignment.assignee_bag:
            assert assignee.assignee_name_text is not None
            assert isinstance(assignee.assignee_address, AssigneeAddress)
            assert assignee.assignee_address.city_name is not None
            assert assignee.assignee_address.geographic_region_code is not None
            assert assignee.assignee_address.postal_code is not None
            assert assignee.assignee_address.address_line_one_text is not None
        for correspondence in assignment.correspondence_address:
            assert correspondence.address_line_one_text is not None
            assert correspondence.address_line_two_text is not None
            assert correspondence.correspondent_name_text is not None


def test_foreign_priority(fixture_dir):
    data = json.loads((fixture_dir / "foreign_priority.json").read_text())
    foreign_priority_bag = data["patentBag"][0]["foreignPriorityBag"]
    for foreign_priority_data in foreign_priority_bag:
        foreign_priority = ForeignPriority(**foreign_priority_data)
        assert foreign_priority.priority_number_text is not None
        assert isinstance(foreign_priority.filing_date, datetime.date)
        assert foreign_priority.country_name is not None


def test_attorney(fixture_dir):
    data = json.loads((fixture_dir / "attorney.json").read_text())
    customer_number = CustomerNumber(**data["patentBag"][0]["recordAttorney"])
    assert customer_number.customer_number == "26161.0000000000"
    assert len(customer_number.attorneys) > 0
    for attorney in customer_number.attorneys:
        assert attorney.active_indicator in ["ACTIVE", "INACTIVE"]
        assert attorney.first_name is not None
        assert attorney.last_name is not None
        assert attorney.registration_number is not None
        assert len(attorney.attorney_address_bag) > 0
        assert len(attorney.telecommunication_address_bag) > 0
        for telecom in attorney.telecommunication_address_bag:
            assert telecom.telecommunication_number is not None
            assert telecom.usage_type_category in ["FAX", "TEL", "EMAIL"]
    assert customer_number.address is not None
    assert customer_number.address.city_name == "MINNEAPOLIS"
    assert customer_number.address.geographic_region_code == "MN"
    assert customer_number.address.country_code == "US"
    assert customer_number.address.postal_code == "55440-1022"
    assert customer_number.address.country_name == "UNITED STATES"
    assert customer_number.address.address_line_one_text == "P.O. BOX 1022"


def test_transactions(fixture_dir):
    data = json.loads((fixture_dir / "transactions.json").read_text())
    transactions = [
        Transaction(**transaction_data)
        for transaction_data in data["patentBag"][0]["transactionContentBag"]
    ]
    for transaction in transactions:
        assert transaction.recorded_date is not None
        assert transaction.transaction_code is not None
        assert transaction.transaction_description is not None
        # Test for specific transaction details based on the provided fixtures
    assert transactions[0].recorded_date == datetime.date.fromisoformat("2016-09-02")
    assert transactions[0].transaction_code == "IEXX"
    assert transactions[0].transaction_description == "Initial Exam Team nn"
    assert transactions[3].recorded_date == datetime.date.fromisoformat("2016-12-01")
    assert transactions[3].transaction_code == "SMAL"
    assert (
        transactions[3].transaction_description
        == "Applicant Has Filed a Verified Statement of Small Entity Status in Compliance with 37 CFR 1.27"
    )
    assert transactions[-1].recorded_date == datetime.date.fromisoformat("2020-10-21")
    assert transactions[-1].transaction_code == "IDSC"
    assert transactions[-1].transaction_description == "Information Disclosure Statement considered"


def test_patent_term_adjustment(fixture_dir):
    data = json.loads((fixture_dir / "adjustment.json").read_text())
    patent_term_adjustment_data = data["patentBag"][0]["patentTermAdjustmentData"]
    patent_term_adjustment = TermAdjustment(**patent_term_adjustment_data)

    assert patent_term_adjustment.applicant_day_delay_quantity == 15
    assert patent_term_adjustment.overlapping_day_quantity == 0
    assert patent_term_adjustment.filing_date == datetime.date.fromisoformat("2018-09-06")
    assert patent_term_adjustment.c_delay_quantity == 0
    assert patent_term_adjustment.adjustment_total_quantity == 0
    assert patent_term_adjustment.b_delay_quantity == 0
    assert patent_term_adjustment.grant_date == datetime.date.fromisoformat("2021-01-26")
    assert patent_term_adjustment.a_delay_quantity == 142
    assert patent_term_adjustment.non_overlapping_day_quantity == 127
    assert patent_term_adjustment.ip_office_day_delay_quantity == 142

    assert len(patent_term_adjustment.history) > 0
    for history_item in patent_term_adjustment.history:
        assert isinstance(history_item, TermAdjustmentHistory)
        assert history_item.applicant_day_delay_quantity >= 0
        assert isinstance(history_item.start_sequence_number, float)
        assert history_item.case_action_description_text is not None
        assert isinstance(history_item.case_action_sequence_number, float)
        assert history_item.action_date is not None


def test_application_biblio(fixture_dir):
    data = json.loads((fixture_dir / "biblio.json").read_text())
    application_biblio_data = data["patentBag"][0]
    application_biblio = USApplicationBiblio(**application_biblio_data)

    assert application_biblio.aia_indicator is True
    assert application_biblio.app_filing_date == datetime.date.fromisoformat("2016-09-02")
    assert len(application_biblio.inventors) > 0
    assert application_biblio.customer_number == 26161
    assert application_biblio.group_art_unit == "3775"
    assert application_biblio.invention_title == "TONGUE RETRACTORS FOR FRENOTOMIES"
    assert len(application_biblio.correspondence_address) > 0
    assert application_biblio.app_conf_num == 7229
    assert application_biblio.atty_docket_num == "29539-0122US1"
    assert application_biblio.appl_id == "15123456"
    assert application_biblio.first_inventor_name == "Marvin Wang"
    assert application_biblio.first_applicant_name == "The General Hospital Corporation"
    assert len(application_biblio.cpc_classifications) > 0
    assert application_biblio.entity_status == "Small"
    assert application_biblio.status == "Abandoned  --  Failure to Respond to an Office Action"
    assert application_biblio.status_date == datetime.date.fromisoformat("2018-10-02")
    assert application_biblio.status_code is None
    assert application_biblio.patent_number is None


def test_application_object(fixture_dir):
    data = json.loads((fixture_dir / "application.json").read_text())
    application_data = data["patentBag"][0]
    application = USApplication(**application_data)

    assert application.aia_indicator is True
    assert application.app_filing_date == datetime.date.fromisoformat("2016-09-02")
    assert len(application.inventors) > 0
    assert application.customer_number == 26161
    assert application.group_art_unit == "3775"
    assert application.invention_title == "TONGUE RETRACTORS FOR FRENOTOMIES"
    assert len(application.correspondence_address) > 0
    assert application.app_conf_num == 7229
    assert application.atty_docket_num == "29539-0122US1"
    assert application.appl_id == "15123456"
    assert application.first_inventor_name == "Marvin Wang"
    assert application.first_applicant_name == "The General Hospital Corporation"
    assert len(application.cpc_classifications) > 0
    assert application.entity_status == "Small"
    assert application.app_type_code == "UTL"
    assert application.national_stage_indicator is False
    assert application.effective_filing_date == datetime.date.fromisoformat("2016-09-02")
    assert len(application.assignments) > 0
    assert len(application.attorneys.attorneys) > 0
    assert len(application.transactions) > 0
    assert len(application.parent_applications) > 0
    assert len(application.child_applications) > 0
    assert application.status == "Abandoned  --  Failure to Respond to an Office Action"
    assert application.status_date == datetime.date.fromisoformat("2018-10-02")
    assert application.status_code == 161
    assert application.patent_number is None


def test_search_model():
    model = SearchRequest(rangeFilters=[{"field": "filingDate"}])
    assert model.rangeFilters[0].field == "filingDate"
    assert model.rangeFilters[0].valueFrom == "1776-07-04"
    assert model.rangeFilters[0].valueTo == datetime.date.today().isoformat()



import re
import typing as tp
import warnings
from hashlib import blake2b
from pathlib import Path

import hishel
import httpcore
import httpx
from hishel._utils import normalized_url

from patent_client import CACHE_DIR
from patent_client.version import __version__

filename_re = re.compile(r'filename="([^"]+)"')


def cache_key_generator(request: httpcore.Request):
    encoded_url = normalized_url(request.url).encode("ascii")
    body = [c for c in request.stream]
    if isinstance(body[0], bytes):
        body = b"".join(body)
    else:
        body = "".join(body).encode("utf-8")
    key = blake2b(digest_size=16)
    key.update(encoded_url)
    key.update(request.method)
    key.update(body)
    return key.hexdigest()


patent_client_transport = hishel.AsyncCacheTransport(
    transport=httpx.AsyncHTTPTransport(
        verify=False,
        http2=True,
        retries=3,
    ),
    storage=hishel.AsyncFileStorage(base_path=CACHE_DIR),
    controller=hishel.Controller(allow_heuristics=True),
)


class PatentClientSession(httpx.AsyncClient):
    _default_user_agent = f"Mozilla/5.0 Python Patent Clientbot/{__version__} (parkerhancock@users.noreply.github.com)"

    def __init__(self, **kwargs):
        kwargs["transport"] = kwargs.get("transport", patent_client_transport)
        headers = kwargs.get("headers", dict())
        headers["User-Agent"] = headers.get("User-Agent", self._default_user_agent)
        kwargs["headers"] = headers
        kwargs["follow_redirects"] = kwargs.get("follow_redirects", True)
        kwargs["timeout"] = kwargs.get("timeout", 60 * 5)
        super().__init__(**kwargs)

    def get_filename(self, url, path, filename, headers):
        if path.is_dir() or None:
            try:
                filename = filename_re.search(headers["Content-Disposition"]).group(1)
            except (AttributeError, KeyError):
                filename = url.split("/")[-1]
            path = path / filename if path else Path.cwd() / filename
        return path

    async def download(
        self,
        url,
        method: str = "GET",
        path: tp.Optional[tp.Union[str, Path]] = None,
        **kwargs,
    ):
        # Ensure we skip the cache for file downloads
        kwargs["extensions"] = kwargs.get("extensions", dict())
        if "force_cache" in kwargs["extensions"]:
            raise ValueError("force_cache is not supported for file downloads")
        kwargs["extensions"]["cache_disabled"] = True
        if isinstance(path, str):
            path = Path(path)
        elif path is None:
            path = Path.cwd()
        if not path.is_dir() and path.exists():
            warnings.warn(
                "File already exists at provided output! Not re-downloading. Please move the file or provide an alternative path to download"
            )
            return path
        async with self.stream(method, url, **kwargs) as response:
            response.raise_for_status()
            path = self.get_filename(url, path, response, response.headers)
            with path.open("wb") as f:
                async for chunk in response.aiter_bytes():
                    f.write(chunk)
        return path



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *         Source File: patent_client/_async/epo/ops/family/api_test.py         *
# ********************************************************************************

import json
from pathlib import Path

from .api import FamilyAsyncApi

fixtures = Path(__file__).parent / "fixtures"


def test_api():
    result = FamilyAsyncApi.get_family("EP1000000A1")
    expected_file = fixtures / "family_api_output.json"
    # expected_file.write_text(result.model_dump_json(indent=2))
    expected = expected_file.read_text()
    assert json.loads(expected) == json.loads(result.model_dump_json())



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *        Source File: patent_client/_async/uspto/global_dossier/api.py         *
# ********************************************************************************

import re
from urllib.parse import urljoin

import lxml.etree as ET

from patent_client._sync.http_client import PatentClientSession

from .model import DocumentList, GlobalDossier

client = PatentClientSession(
    headers={
        "Authorization": "OQmPwAN1QD4OXe25jpmMD27zmnM21gIL0lg85G6j",
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/21.1.1.164 Safari/602.1.507",
    }
)


class GlobalDossierApi:
    def __init__(self):
        self.api_base_url = None

    def get_base_url(self):
        if self.api_base_url:
            return self.api_base_url
        base_url = "https://globaldossier.uspto.gov"
        response = client.get(base_url)
        tree = ET.HTML(response.text)
        scripts = [urljoin(base_url, x) for x in tree.xpath("//script/@src")]
        main_script = next(filter(lambda x: "main" in x, scripts))
        response = client.get(main_script)
        api_url_re = re.compile(r'apiURL:\s*"([^"]+)"')
        match = api_url_re.search(response.text)
        self.api_base_url = match.group(1) if match else None
        return self.api_base_url

    def get_file(self, doc_number, type_code="application", office_code="US"):
        url = (
            f"{self.get_base_url()}/patent-family/svc/family/{type_code}/{office_code}/{doc_number}"
        )
        _ = client.options(url)
        response = client.get(url)
        response.raise_for_status()
        return GlobalDossier.model_validate_json(response.content)

    def get_doc_list(self, country, doc_number, kind_code):
        url = f"{self.get_base_url()}/doc-list/svc/doclist/{country}/{doc_number}/{kind_code}"
        response = client.get(url)
        response.raise_for_status()
        return DocumentList.model_validate_json(response.content)

    def get_document(self, country, doc_number, document_id, out_path):
        url = f"{self.get_base_url()}/doc-content/svc/doccontent/{country}/{doc_number}/{document_id}/1/PDF"
        return client.download(url, path=out_path)



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *        Source File: patent_client/_async/epo/ops/family/model_test.py        *
# ********************************************************************************

import json
from pathlib import Path

from patent_client.util.test import compare_dicts

from .model import Family

fixtures = Path(__file__).parent / "fixtures"


def test_model():
    result = Family.objects.get("EP1000000A1")
    expected_file = fixtures / "family_model_output.json"
    # expected_file.write_text(result.model_dump_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.model_dump_json()), expected)



from patent_client.util.manager import AsyncManager

from .api import FamilyAsyncApi
from .model import Family
from .schema import FamilySchema


class FamilyManager(AsyncManager[Family]):
    __schema__ = FamilySchema

    async def get(self, doc_number):
        return await FamilyAsyncApi.get_family(doc_number, doc_type="publication", format="docdb")



# A simple separate namespace for the USPTO's Open Data Portal.

from patent_client._async.uspto.odp.model import USApplication, USApplicationBiblio

__all__ = ["USApplication", "USApplicationBiblio"]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *            Source File: patent_client/_async/epo/ops/__init__.py             *
# ********************************************************************************

from .family.model import Family
from .legal.model import Legal
from .published.model import Images, Inpadoc

__api_name__ = "EPO Open Patent Services"
__all__ = ["Family", "Legal", "Images", "Inpadoc", "session"]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *    Source File: patent_client/_async/epo/ops/legal/national_codes_test.py    *
# ********************************************************************************

import datetime

from ..legal import national_codes


def stub_date():
    return datetime.datetime(2022, 8, 1).date()


def test_legal_codes():
    dt = national_codes.current_date
    national_codes.current_date = stub_date
    national_codes.current_date = dt


def test_epo_website_parsing():
    sp_date, sp_url = national_codes.get_spreadsheet_from_epo_website()
    assert isinstance(sp_date, datetime.date)
    assert isinstance(sp_url, str)



import datetime
from enum import Enum
from typing import Any, List, Optional

from async_property import async_property
from async_property.base import AsyncPropertyDescriptor
from pydantic import AliasPath, BeforeValidator, ConfigDict, Field, model_validator
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated

from patent_client.util.pydantic_util import BaseModel


class BaseODPModel(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel, ignored_types=(AsyncPropertyDescriptor,))


# Common


class Address(BaseODPModel):
    city_name: Optional[str] = Field(alias="cityName", default=None)
    geographic_region_name: Optional[str] = Field(alias="geographicRegionName", default=None)
    geographic_region_code: Optional[str] = Field(alias="geographicRegionCode", default=None)
    country_code: Optional[str] = Field(alias="countryCode", default=None)
    postal_code: Optional[str] = Field(alias="postalCode", default=None)
    country_name: Optional[str] = Field(alias="countryName", default=None)
    address_line_one_text: Optional[str] = Field(alias="addressLineOneText", default=None)
    address_line_two_text: Optional[str] = Field(alias="addressLineTwoText", default=None)
    name_line_one_text: Optional[str] = Field(alias="nameLineOneText", default=None)
    name_line_two_text: Optional[str] = Field(alias="nameLineTwoText", default=None)
    postal_address_category: Optional[str] = Field(alias="postalAddressCategory", default=None)
    correspondent_name_text: Optional[str] = Field(alias="correspondentNameText", default=None)


# Continuity


class Relationship(BaseODPModel):
    application_status_code: Optional[int] = Field(default=None)
    claim_type_code: Optional[str] = Field(alias="claimParentageTypeCode", default=None)
    filing_date: Optional[datetime.date] = Field(default=None)
    application_status_description: Optional[str] = Field(
        alias="applicationStatusDescriptionText", default=None
    )
    claim_type_description: Optional[str] = Field(
        alias="claimParentageTypeCodeDescription", default=None
    )
    parent_application_id: Optional[str] = Field(alias="parentApplicationNumberText", default=None)
    child_application_id: Optional[str] = Field(alias="childApplicationNumberText", default=None)


class Continuity(BaseODPModel):
    count: Optional[int] = Field(default=None)
    request_identifier: Optional[str] = Field(default=None)
    parent_continuity: Optional[list[Relationship]] = Field(
        alias=AliasPath(["patentBag", 0, "continuityBag", "parentContinuityBag"]),
        default_factory=list,
    )
    child_continuity: Optional[list[Relationship]] = Field(
        alias=AliasPath(["patentBag", 0, "continuityBag", "childContinuityBag"]),
        default_factory=list,
    )


# Documents


class DownloadOption(BaseODPModel):
    mime_type_identifier: Optional[str] = Field(default=None)
    download_url: Optional[str] = Field(default=None)
    pages: Optional[int] = Field(alias="pageTotalQuantity", default=None)


class Document(BaseODPModel):
    appl_id: Optional[str] = Field(alias="applicationNumberText", default=None)
    mail_date: Optional[datetime.datetime] = Field(alias="officialDate", default=None)
    document_identifier: Optional[str] = Field(alias="documentIdentifier", default=None)
    document_code: Optional[str] = Field(alias="documentCode", default=None)
    document_code_description: Optional[str] = Field(
        alias="documentCodeDescriptionText", default=None
    )
    direction_category: Optional[str] = Field(alias="directionCategory", default=None)
    download_option_bag: list[dict] = Field(alias="downloadOptionBag", default_factory=list)

    async def download(self, type="PDF", out_path=None):
        from .manager import api

        try:
            url = next(u for u in self.download_option_bag if u["mimeTypeIdentifier"] == type)[
                "downloadUrl"
            ]
        except StopIteration:
            raise ValueError(f"No download URL found for this document type: {type}")
        if out_path is None:
            out_path = f"{self.appl_id} - {self.mail_date.date()} - {self.document_code} - {self.document_code_description}.{type.lower()}".replace(
                "/", "-"
            )
        return await api.client.download(url, "GET", path=out_path)


# Assignment


class Assignor(BaseODPModel):
    execution_date: Optional[datetime.date] = Field(alias="executionDate", default=None)
    assignor_name: Optional[str] = Field(alias="assignorName", default=None)


class AssigneeAddress(BaseODPModel):
    city_name: Optional[str] = Field(alias="cityName", default=None)
    geographic_region_code: Optional[str] = Field(alias="geographicRegionCode", default=None)
    postal_code: Optional[str] = Field(alias="postalCode", default=None)
    address_line_one_text: Optional[str] = Field(alias="addressLineOneText", default=None)


class Assignee(BaseODPModel):
    assignee_address: Optional[AssigneeAddress] = Field(alias="assigneeAddress", default=None)
    assignee_name_text: Optional[str] = Field(alias="assigneeNameText", default=None)


class Assignment(BaseODPModel):
    assignment_received_date: Optional[datetime.date] = Field(
        alias="assignmentReceivedDate", default=None
    )
    frame_number: Optional[int] = Field(alias="frameNumber", default=None)
    page_number: Optional[int] = Field(alias="pageNumber", default=None)
    reel_number_frame_number: Optional[str] = Field(alias="reelNumber/frameNumber", default=None)
    assignment_recorded_date: Optional[datetime.date] = Field(
        alias="assignmentRecordedDate", default=None
    )
    conveyance_text: Optional[str] = Field(alias="conveyanceText", default=None)
    assignment_mailed_date: Optional[datetime.date] = Field(
        alias="assignmentMailedDate", default=None
    )
    reel_number: Optional[int] = Field(alias="reelNumber", default=None)
    assignor_bag: list[Assignor] = Field(alias="assignorBag", default_factory=list)
    assignee_bag: list[Assignee] = Field(alias="assigneeBag", default_factory=list)
    correspondence_address: list[Address] = Field(
        alias="correspondenceAddress", default_factory=list
    )


# Foreign Priority


class ForeignPriority(BaseODPModel):
    priority_number_text: Optional[str] = Field(alias="priorityNumberText", default=None)
    filing_date: Optional[datetime.date] = Field(alias="filingDate", default=None)
    country_name: Optional[str] = Field(alias="countryName", default=None)


# Attorney


class TelecommunicationAddress(BaseODPModel):
    telecommunication_number: Optional[str] = Field(alias="telecommunicationNumber", default=None)
    usage_type_category: Optional[str] = Field(alias="usageTypeCategory", default=None)


class Attorney(BaseODPModel):
    active_indicator: Optional[str] = Field(alias="activeIndicator", default=None)
    first_name: Optional[str] = Field(alias="firstName", default=None)
    last_name: Optional[str] = Field(alias="lastName", default=None)
    registration_number: Optional[str] = Field(alias="registrationNumber", default=None)
    attorney_address_bag: list[Address] = Field(alias="attorneyAddressBag", default_factory=list)
    telecommunication_address_bag: list[TelecommunicationAddress] = Field(
        alias="telecommunicationAddressBag", default_factory=list
    )
    registered_practitioner_category: Optional[str] = Field(
        alias="registeredPractitionerCategory", default=None
    )
    name_suffix: Optional[str] = Field(alias="nameSuffix", default=None)


class CustomerNumber(BaseODPModel):
    attorneys: list[Attorney] = Field(alias="attorneyBag", default_factory=list)
    customer_number: Optional[str] = Field(
        alias=AliasPath("customerNumber", "patronIdentifier"), default=None
    )
    address: Optional[Address] = Field(
        alias=AliasPath("customerNumber", "powerOfAttorneyAddressBag", 0), default=None
    )


# Transactions


class Transaction(BaseODPModel):
    recorded_date: Optional[datetime.date] = Field(alias="recordedDate", default=None)
    transaction_code: Optional[str] = Field(alias="caseActionCode", default=None)
    transaction_description: Optional[str] = Field(alias="caseActionDescriptionText", default=None)


# Adjustment Data
class TermAdjustmentHistory(BaseODPModel):
    applicant_day_delay_quantity: Optional[int] = Field(
        alias="applicantDayDelayQuantity", default=None
    )
    start_sequence_number: Optional[float] = Field(alias="startSequenceNumber", default=None)
    case_action_description_text: Optional[str] = Field(
        alias="caseActionDescriptionText", default=None
    )
    case_action_sequence_number: Optional[float] = Field(
        alias="caseActionSequenceNumber", default=None
    )
    action_date: Optional[datetime.date] = Field(alias="actionDate", default=None)


class TermAdjustment(BaseODPModel):
    applicant_day_delay_quantity: Optional[int] = Field(
        alias="applicantDayDelayQuantity", default=None
    )
    overlapping_day_quantity: Optional[int] = Field(alias="overlappingDayQuantity", default=None)
    filing_date: Optional[datetime.date] = Field(alias="filingDate", default=None)
    c_delay_quantity: Optional[int] = Field(alias="cDelayQuantity", default=None)
    adjustment_total_quantity: Optional[int] = Field(alias="adjustmentTotalQuantity", default=None)
    b_delay_quantity: Optional[int] = Field(alias="bDelayQuantity", default=None)
    grant_date: Optional[datetime.date] = Field(alias="grantDate", default=None)
    a_delay_quantity: Optional[int] = Field(alias="aDelayQuantity", default=None)
    non_overlapping_day_quantity: Optional[int] = Field(
        alias="nonOverlappingDayQuantity", default=None
    )
    ip_office_day_delay_quantity: Optional[int] = Field(
        alias="ipOfficeDayDelayQuantity", default=None
    )
    history: Optional[list[TermAdjustmentHistory]] = Field(
        alias="patentTermAdjustmentHistoryDataBag", default_factory=list
    )


# Application Object

YNBool = Annotated[bool, BeforeValidator(lambda v: v == "Y")]


class Inventor(BaseODPModel):
    first_name: Optional[str] = Field(alias="firstName", default=None)
    last_name: Optional[str] = Field(alias="lastName", default=None)
    full_name: Optional[str] = Field(alias="inventorNameText", default=None)
    addresses: list[Address] = Field(alias="correspondenceAddressBag", default_factory=list)


class Applicant(BaseODPModel):
    applicant_name: Optional[str] = Field(alias="applicantNameText", default=None)
    addresses: list[Address] = Field(alias="correspondenceAddressBag", default_factory=list)
    app_status_code: Optional[int] = Field(alias="applicationStatusCode", default=None)
    app_status: Optional[str] = Field(alias="applicationStatusDescriptionText", default=None)


class USApplicationBiblio(BaseODPModel):
    aia_indicator: Optional[YNBool] = Field(alias="firstInventorToFileIndicator", default=None)
    app_filing_date: Optional[datetime.date] = Field(alias="filingDate", default=None)
    inventors: list[Inventor] = Field(alias="inventorBag", default_factory=list)
    customer_number: Optional[int] = Field(alias="customerNumber", default=None)
    group_art_unit: Optional[str] = Field(alias="groupArtUnitNumber", default=None)
    invention_title: Optional[str] = Field(alias="inventionTitle", default=None)
    correspondence_address: list[Address] = Field(
        alias="correspondenceAddressBag", default_factory=list
    )
    app_conf_num: Optional[int] = Field(alias="applicationConfirmationNumber", default=None)
    atty_docket_num: Optional[str] = Field(alias="docketNumber", default=None)
    appl_id: Optional[str] = Field(alias="applicationNumberText", default=None)
    first_inventor_name: Optional[str] = Field(alias="firstInventorName", default=None)
    first_applicant_name: Optional[str] = Field(alias="firstApplicantName", default=None)
    cpc_classifications: list[str] = Field(alias="cpcClassificationBag", default_factory=list)
    entity_status: Optional[str] = Field(alias="businessEntityStatusCategory", default=None)
    app_early_pub_number: Optional[str] = Field(alias="earliestPublicationNumber", default=None)
    app_early_pub_date: Optional[datetime.date] = Field(
        alias="earliestPublicationDate", default=None
    )
    patent_number: Optional[str] = Field(alias="patentNumber", default=None)
    grant_date: Optional[datetime.date] = Field(alias="grantDate", default=None)
    status: Optional[str] = Field(alias="applicationStatusDescriptionText", default=None)
    status_date: Optional[datetime.date] = Field(alias="applicationStatusDate", default=None)
    status_code: Optional[int] = Field(alias="applicationStatusCode", default=None)

    @async_property
    async def bibliographic_data(self) -> "USApplicationBiblio":
        return await self._get_model(".model.USApplicationBiblio").objects.get(appl_id=self.appl_id)

    @async_property
    async def application(self) -> "USApplication":
        return await self._get_model(".model.USApplication").objects.get(appl_id=self.appl_id)

    @async_property
    async def continuity(self) -> Continuity:
        return await self._get_model(".model.Continuity").objects.get(appl_id=self.appl_id)

    @async_property
    async def documents(self) -> list[Document]:
        return self._get_model(".model.Document").objects.filter(appl_id=self.appl_id)

    @async_property
    async def term_adjustment(self) -> TermAdjustment:
        return self._get_model(".model.TermAdjustment").objects.filter(appl_id=self.appl_id)

    @async_property
    async def assignments(self) -> list[Assignment]:
        return self._get_model(".model.Assignment").objects.filter(appl_id=self.appl_id)

    @async_property
    async def customer_number(self) -> CustomerNumber:
        return self._get_model(".model.CustomerNumber").objects.filter(appl_id=self.appl_id)

    @async_property
    async def foreign_priority(self) -> ForeignPriority:
        return self._get_model(".model.ForeignPriority").objects.filter(appl_id=self.appl_id)

    @async_property
    async def transactions(self) -> list[Transaction]:
        return self._get_model(".model.Transaction").objects.filter(appl_id=self.appl_id)

    # Aliases

    @async_property
    async def biblio(self) -> "USApplicationBiblio":
        return await self.bibliographic_data

    @async_property
    async def app(self) -> "USApplication":
        return await self.application

    @async_property
    async def docs(self) -> list[Document]:
        return await self.documents


class USApplication(BaseODPModel):
    aia_indicator: Optional[YNBool] = Field(alias="firstInventorToFileIndicator", default=None)
    app_filing_date: Optional[datetime.date] = Field(alias="filingDate", default=None)
    inventors: list[Inventor] = Field(alias="inventorBag", default_factory=list)
    customer_number: Optional[int] = Field(alias="customerNumber", default=None)
    group_art_unit: Optional[str] = Field(alias="groupArtUnitNumber", default=None)
    invention_title: Optional[str] = Field(alias="inventionTitle", default=None)
    correspondence_address: list[Address] = Field(
        alias="correspondenceAddressBag", default_factory=list
    )
    app_conf_num: Optional[int] = Field(alias="applicationConfirmationNumber", default=None)
    atty_docket_num: Optional[str] = Field(alias="docketNumber", default=None)
    appl_id: Optional[str] = Field(alias="applicationNumberText", default=None)
    first_inventor_name: Optional[str] = Field(alias="firstInventorName", default=None)
    first_applicant_name: Optional[str] = Field(alias="firstApplicantName", default=None)
    cpc_classifications: list[str] = Field(alias="cpcClassificationBag", default_factory=list)
    entity_status: Optional[str] = Field(alias="businessEntityStatusCategory", default=None)
    app_early_pub_number: Optional[str] = Field(alias="earliestPublicationNumber", default=None)
    app_early_pub_date: Optional[datetime.date] = Field(
        alias="earliestPublicationDate", default=None
    )
    patent_number: Optional[str] = Field(alias="patentNumber", default=None)
    grant_date: Optional[datetime.date] = Field(alias="grantDate", default=None)

    app_type_code: Optional[str] = Field(alias="applicationTypeCode", default=None)
    national_stage_indicator: Optional[YNBool] = Field(alias="nationalStageIndicator", default=None)

    status: Optional[str] = Field(alias="applicationStatusDescriptionText", default=None)
    status_date: Optional[datetime.date] = Field(alias="applicationStatusDate", default=None)
    status_code: Optional[int] = Field(alias="applicationStatusCode", default=None)

    effective_filing_date: Optional[datetime.date] = Field(
        alias="effectiveFilingDate", default=None
    )
    cls_sub_cls: Optional[str] = Field(alias="class/subclass", default=None)
    assignments: list[Assignment] = Field(alias="assignmentBag", default_factory=list)
    attorneys: Optional[CustomerNumber] = Field(alias="recordAttorney", default=None)
    transactions: list[Transaction] = Field(alias="transactionContentBag", default_factory=list)
    parent_applications: Optional[list[Relationship]] = Field(
        alias=AliasPath("continuityBag", "parentContinuityBag"), default_factory=list
    )
    child_applications: Optional[list[Relationship]] = Field(
        alias=AliasPath("continuityBag", "childContinuityBag"), default_factory=list
    )
    patent_term_adjustment: Optional[TermAdjustment] = Field(
        alias="patentTermAdjustmentData", default=None
    )

    @model_validator(mode="before")
    @classmethod
    def _validate_patent_term_adjustment(cls, v):
        if "patentTermAdjustmentData" in v and v["patentTermAdjustmentData"] == dict():
            v["patentTermAdjustmentData"] = None
        return v


## RESPONSE Models


class SearchResult(BaseODPModel):
    filing_date: Optional[datetime.date] = Field(default=None)
    appl_id: Optional[str] = Field(alias="applicationNumberText", default=None)
    invention_title: Optional[str] = Field(alias="inventionTitle", default=None)
    filing_date: Optional[datetime.date] = Field(default=None)
    patent_number: Optional[str] = Field(alias="patentNumber", default=None)


class SearchResponse(BaseODPModel):
    count: Optional[int] = Field(default=None)
    results: list[SearchResult] = Field(alias="patentBag", default_factory=list)
    request_id: Optional[str] = Field(alias="requestIdentifier", default=None)


## Request Models


class Filter(BaseModel):
    name: Optional[str] = Field(default=None)
    value: Optional[List[str]] = Field(default_factory=list)


class Range(BaseModel):
    field: Optional[str] = Field(examples=["grantDate"], default=None)
    valueFrom: Optional[str] = Field(examples=["2020-01-01"], default=None)
    valueTo: Optional[str] = Field(examples=["2020-12-31"], default=None)

    @model_validator(mode="before")
    @classmethod
    def add_default_dates(cls, data: Any) -> Any:
        if data.get("valueFrom") is None:
            data["valueFrom"] = "1776-07-04"
        if data.get("valueTo") is None:
            data["valueTo"] = datetime.date.today().isoformat()
        return data


class SortOrder(str, Enum):
    asc = "asc"
    desc = "desc"
    Asc = "Asc"
    Desc = "Desc"


class Sort(BaseModel):
    field: Optional[str] = Field(examples=["grantDate"], default=None)
    order: Optional[SortOrder] = Field(examples=[SortOrder.desc], default=None)


class Pagination(BaseModel):
    offset: Optional[int] = Field(examples=[0], default=0, ge=0)
    limit: Optional[int] = Field(examples=[25], default=25, ge=1)


class SearchRequest(BaseModel):
    q: Optional[str] = Field(default="")
    filters: Optional[List[Filter]] = Field(default_factory=list)
    rangeFilters: Optional[List[Range]] = Field(default_factory=list)
    sort: Optional[List[Sort]] = Field(default_factory=list)
    fields: Optional[List[str]] = Field(default_factory=list)
    pagination: Optional[Pagination] = Field(default=None)
    facets: Optional[List[str]] = Field(default_factory=list)


class SearchGetRequest(BaseModel):
    q: Optional[str] = Field(default="")
    sort: Optional[str] = Field(default="filingDate")
    fields: Optional[str] = Field(default="")
    offset: Optional[int] = Field(default=0)
    limit: Optional[int] = Field(default=25)



# flake8: noqa
# nopycln: file
import time

import patent_client.patches  # noqa # Run patching code

from .version import __version__  # noqa

start = time.time()

from pathlib import Path

from .settings import Settings

SETTINGS = Settings()

import logging
import logging.handlers

# Revert base directory to local if there's an access problem

BASE_DIR = SETTINGS.base_dir
try:
    BASE_DIR.mkdir(exist_ok=True, parents=True)
except OSError:
    BASE_DIR = Path(__file__).parent.parent.parent / "_build"
    BASE_DIR.mkdir(exist_ok=True, parents=True)
    SETTINGS.DEFAULT.BASE_DIR = str(BASE_DIR)

LOG_FILENAME = BASE_DIR / SETTINGS.log_file
CACHE_DIR = BASE_DIR / "cache"
CACHE_DIR.mkdir(exist_ok=True, parents=True)

# Set up a specific logger with our desired output level
logger = logging.getLogger(__name__)
logger.setLevel(SETTINGS.log_level)


# Add the log message handler to the logger
handler = logging.FileHandler(LOG_FILENAME)
handler.setFormatter(logging.Formatter("%(asctime)s:%(levelname)s:%(name)s:%(message)s"))
logger.addHandler(handler)

logger.info(f"Starting Patent Client with log level {SETTINGS.log_level}")

try:
    from ._sync import *
except ImportError:
    pass

elapsed = time.time() - start
logger.info(f"Startup Complete!, took {elapsed:.3f} seconds")

cache_dir = Path("~/.patent_client/cache").expanduser()
cache_dir.parent.mkdir(exist_ok=True, parents=True)

__all__ = [
    "Inpadoc",
    "Assignment",
    "PublicSearch",
    "PublicSearchDocument",
    "Patent",
    "PatentBiblio",
    "PublishedApplication",
    "PublishedApplicationBiblio",
    "USApplication",
    "PtabDecision",
    "PtabDocument",
    "PtabProceeding",
    "GlobalDossier",
]



from .model import (
    Patent,
    PatentBiblio,
    PublicSearchBiblio,
    PublicSearchDocument,
    PublishedApplication,
    PublishedApplicationBiblio,
)

__all__ = [
    "PatentBiblio",
    "Patent",
    "PublishedApplicationBiblio",
    "PublishedApplication",
    "PublicSearchBiblio",
    "PublicSearchDocument",
]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *              Source File: patent_client/_async/epo/ops/util.py               *
# ********************************************************************************

from typing import TYPE_CHECKING, List, Optional

from async_property.base import AsyncPropertyDescriptor
from pydantic import ConfigDict, computed_field, model_validator
from yankee.xml.schema import Schema as XmlSchema

from patent_client.util.pydantic_util import BaseModel

if TYPE_CHECKING:
    from .family.model import FamilyMember
    from .legal.model import LegalEvent
    from .published.model import Claims, Images, InpadocBiblio


class Schema(XmlSchema):
    class Meta:
        namespaces = {
            "ops": "http://ops.epo.org",
            "epo": "http://www.epo.org/exchange",
            "ft": "http://www.epo.org/fulltext",
        }


class EpoBaseModel(BaseModel):
    model_config = ConfigDict(ignored_types=(AsyncPropertyDescriptor,))
    __schema__: Optional[XmlSchema] = None

    @model_validator(mode="before")
    @classmethod
    def xml_convert(cls, values):
        if isinstance(values, (str, bytes)):
            return cls.__schema__.load(values).to_dict()
        return values


class InpadocModel(EpoBaseModel):
    @computed_field
    @property
    def docdb_number(self) -> str:
        num = self.publication_reference_docdb
        if num.kind:
            return f"{num.country}{num.number}.{num.kind}"
        return f"{num.country}{num.number}"

    @property
    def biblio(self) -> "InpadocBiblio":
        return self._get_model(
            ".published.model.InpadocBiblio", base_class=InpadocModel
        ).objects.get(self.docdb_number)

    @property
    def images(self) -> "Images":
        return self._get_model(".published.model.Images", base_class=InpadocModel).objects.get(
            self.docdb_number
        )

    @property
    def description(self) -> Optional[str]:
        return (
            self._get_model(".published.model.Description", base_class=InpadocModel).objects.get(
                self.docdb_number
            )
        ).description

    @property
    def claims(self) -> "Claims":
        return self._get_model(".published.model.Claims", base_class=InpadocModel).objects.get(
            self.docdb_number
        )

    @property
    def legal(self) -> List["LegalEvent"]:
        return (
            self._get_model(".legal.model.Legal", base_class=InpadocModel).objects.get(
                self.docdb_number
            )
        ).events

    @property
    def family(
        self,
    ) -> List["FamilyMember"]:
        return (
            self._get_model(".family.model.Family", base_class=InpadocModel).objects.get(
                self.docdb_number
            )
        ).family_members

    def download(self, path: str = "."):
        images = self.images
        return images.full_document.download(path)



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *          Source File: patent_client/_async/epo/ops/family/model.py           *
# ********************************************************************************

from typing import List, Optional

from pydantic import Field

from ..number_service.model import DocumentId
from ..util import EpoBaseModel
from .schema import FamilySchema


class PriorityClaim(EpoBaseModel):
    application_number: Optional[str] = None
    application_reference: Optional[DocumentId] = None
    sequence: Optional[int] = None
    kind: Optional[str] = None
    active: Optional[bool] = None


class FamilyMember(EpoBaseModel):
    publication_number: Optional[str] = None
    application_number: Optional[str] = None
    family_id: Optional[str] = None
    publication_reference: list = Field(default_factory=list)
    application_reference: list = Field(default_factory=list)
    priority_claims: List[PriorityClaim] = Field(default_factory=list)

    @property
    def docdb_number(self):
        return self.publication_number


class Family(EpoBaseModel):
    __schema__ = FamilySchema()
    publication_reference: Optional[DocumentId] = None
    num_records: Optional[int] = None
    publication_number: Optional[str] = None
    family_members: List[FamilyMember] = Field(default_factory=list)



from pathlib import Path

import pytest

from .api import AssignmentApi


@pytest.mark.asyncio
async def test_alookup():
    query = "test_query"
    filter = "OwnerName"
    sort = "ExecutionDate+desc"
    rows = 5
    start = 0
    assignments = await AssignmentApi.lookup(
        query=query, filter=filter, rows=rows, start=start, sort=sort
    )
    assert assignments is not None, "Expected to get a response"
    assert (
        len(assignments) <= rows
    ), "Expected number of assignments to be less than or equal to requested rows"


@pytest.mark.asyncio
async def test_download_pdf(tmp_path: Path):
    reel = "12345"
    frame = "67890"
    path = await AssignmentApi.download_pdf(reel=reel, frame=frame, path=tmp_path)
    assert path.exists(), "Expected the PDF file to exist"
    assert path.is_file(), "Expected the path to be a file"
    assert path.suffix == ".pdf", "Expected the file to be a PDF"



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *      Source File: patent_client/_async/uspto/public_search/api_test.py       *
# ********************************************************************************

from .api import PublicSearchApi


def test_simple_search():
    api = PublicSearchApi()
    results = api.run_query(
        query='("6013599").pn.',
    )
    assert results.num_found == 1
    assert results.docs[0].publication_number == "6013599"



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *     Source File: patent_client/_async/epo/ops/published/manager_test.py      *
# ********************************************************************************

from .model.search import Inpadoc


class TestPublished:
    def test_inpadoc_manager(self):
        result = Inpadoc.objects.filter(applicant="Microsoft")
        assert result.count() > 20
        countries = [c for c in result.limit(20).values_list("country", flat=True)]
        assert sum(1 for c in countries if c == "US") >= 1

    def test_get_biblio_from_result(self):
        doc = Inpadoc.objects.filter(applicant="Google").first()
        result = doc.biblio
        assert len(result.titles) > 0

    def test_get_claims_from_result(self):
        result = Inpadoc.objects.get("WO2009085664A2")
        claims = result.claims
        assert len(claims.claims) == 20
        assert len(claims.claim_text) == 4830

    def test_get_description_from_result(self):
        result = Inpadoc.objects.get("WO2009085664A2")
        description = result.description
        assert len(description) == 47955

    def test_get_family_from_result(self):
        result = Inpadoc.objects.get("WO2009085664A2")
        family = result.family
        assert len(family) >= 20

    def test_get_biblio_from_wo(self):
        result = Inpadoc.objects.get("WO2009085664A2")
        biblio = result.biblio
        assert biblio.abstract is not None

    def test_can_index_inpadoc_result(self):
        result = Inpadoc.objects.filter(applicant="Tesla")
        first = result.first()
        second = result.offset(1).first()
        assert first != second

    def test_can_handle_single_item_ipc_classes(self):
        result = Inpadoc.objects.get("WO2020081771")
        biblio = result.biblio
        assert biblio.intl_class is not None

    def test_issue_41(self):
        result = Inpadoc.objects.get("JP2005533465A")
        biblio = result.biblio
        assert biblio.title is None

    def test_document_downloading(self, tmpdir):
        result = Inpadoc.objects.get("WO2010050748A2")
        result.download(tmpdir)
        assert len(tmpdir.listdir()) > 0



from pathlib import Path

import pytest

from .api import GlobalDossierApi
from .model import DocumentList, GlobalDossier


@pytest.mark.asyncio
async def test_get_file():
    api = GlobalDossierApi()
    doc_number = "16740760"
    type_code = "application"
    office_code = "US"
    response = await api.get_file(doc_number, type_code, office_code)
    assert isinstance(response, GlobalDossier)
    assert response.country == "US"
    assert response.id == "16740760"


@pytest.mark.asyncio
async def test_get_doc_list():
    api = GlobalDossierApi()
    country = "US"
    doc_number = "16740760"
    kind_code = "A"
    response = await api.get_doc_list(country, doc_number, kind_code)
    assert isinstance(response, DocumentList)
    assert len(response.docs) > 0


@pytest.mark.asyncio
async def test_get_document(tmp_path: Path):
    api = GlobalDossierApi()
    country = "US"
    doc_number = "201816123456.A"
    document_id = "KJM2MWIEDFLYX10"
    out_path = tmp_path / "test_document.pdf"
    await api.get_document(country, doc_number, document_id, out_path)
    assert out_path.exists()
    assert out_path.stat().st_size > 0



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *            Source File: patent_client/_async/uspto/odp/manager.py            *
# ********************************************************************************

import typing as tp

from patent_client.util.manager import Manager
from patent_client.util.request_util import get_start_and_row_count

from .api import ODPApi
from .model import SearchRequest, USApplication, USApplicationBiblio
from .query import create_post_search_obj

if tp.TYPE_CHECKING:
    from .model import (
        Assignment,
        Continuity,
        CustomerNumber,
        Document,
        ForeignPriority,
        SearchResult,
        TermAdjustment,
        Transaction,
        USApplication,
        USApplicationBiblio,
    )
api = ODPApi()


class USApplicationManager(Manager):
    default_filter = "appl_id"
    default_fields = ["applicationNumberText"]
    response_model = USApplication

    def count(self):
        return (api.post_search(self._create_search_obj(fields=["applicationNumberText"])))["count"]

    def _get_results(self) -> tp.Iterator["SearchResult"]:
        query_obj = self._create_search_obj()
        for start, rows in get_start_and_row_count(self.config.limit):
            page_query = query_obj.model_dump()
            page_query["pagination"] = {"offset": start, "limit": rows}
            page_query_obj = SearchRequest(**page_query)
            for result in (api.post_search(page_query_obj))["patentBag"]:
                app_id = result["applicationNumberText"]
                app = api.get_application_data(app_id)
                yield app

    def _create_search_obj(self, fields: tp.Optional[tp.List[str]] = None):
        if fields is None:
            fields = self.default_fields
        if "query" in self.config.filter:
            return SearchRequest(**self.config.filter["query"][0], fields=fields)
        elif "q" in self.config.filter:
            return SearchRequest(q=self.config.filter["q"][0], fields=fields)
        else:
            return create_post_search_obj(self.config, fields=fields)

    def get(self, *args, **kwargs):
        if len(args) == 1 and not kwargs:
            return api.get_application_data(args[0])
        return super().get(*args, **kwargs)


class USApplicationBiblioManager(USApplicationManager):
    default_filter = "appl_id"
    default_fields = [
        "firstInventorToFileIndicator",
        "filingDate",
        "inventorBag",
        "customerNumber",
        "groupArtUnitNumber",
        "inventionTitle",
        "correspondenceAddressBag",
        "applicationConfirmationNumber",
        "docketNumber",
        "applicationNumberText",
        "firstInventorName",
        "firstApplicantName",
        "cpcClassificationBag",
        "businessEntityStatusCategory",
        "earliestPublicationNumber",
    ]
    response_model = USApplicationBiblio

    def _get_results(self) -> tp.Iterator["SearchResult"]:
        query_obj = self._create_search_obj(fields=self.default_fields)
        for start, rows in get_start_and_row_count(self.config.limit):
            page_query = query_obj.model_dump()
            page_query["pagination"] = {"offset": start, "limit": rows}
            page_query_obj = SearchRequest(**page_query)
            for result in (api.post_search(page_query_obj))["patentBag"]:
                yield self.response_model(**result)

    def get(self, *args, **kwargs):
        if len(args) == 1 and not kwargs:
            return api.get_application_biblio_data(args[0])
        return super().get(*args, **kwargs)


class AttributeManager(Manager):
    def filter(self, *args, **kwargs):
        raise NotImplementedError("Filtering attributes is not supported")

    def get(self, *args, **kwargs):
        raise NotImplementedError("Getting attributes is not supported")

    def limit(self, *args, **kwargs):
        raise NotImplementedError("Limit is not supported")

    def offset(self, *args, **kwargs):
        raise NotImplementedError("Offset is not supported")


class ContinuityManager(AttributeManager):
    def get(self, appl_id: str) -> "Continuity":
        return api.get_continuity_data(appl_id)


class DocumentManager(Manager):
    default_filter = "appl_id"

    def count(self):
        return len(api.get_documents(self.config.filter["appl_id"][0]))

    def _get_results(self) -> tp.Iterator["Document"]:
        for doc in api.get_documents(self.config.filter["appl_id"][0]):
            yield doc


class TermAdjustmentManager(Manager):
    default_filter = "appl_id"

    def _get_results(self) -> "TermAdjustment":
        return api.get_term_adjustments(self.config.filter["appl_id"][0])


class AssignmentManager(Manager):
    default_filter = "appl_id"

    def _get_results(self) -> tp.Iterator["Assignment"]:
        for doc in api.get_assignments(self.config.filter["appl_id"][0]):
            yield doc

    def count(self):
        return len(api.get_assignments(self.config.filter["appl_id"][0]))


class CustomerNumberManager(Manager):
    default_filter = "appl_id"

    def _get_results(self) -> "CustomerNumber":
        return api.get_customer_numbers(self.config.filter["appl_id"][0])


class ForeignPriorityManager(Manager):
    default_filter = "appl_id"

    def _get_results(self) -> "ForeignPriority":
        for doc in api.get_foreign_priority_data(self.config.filter["appl_id"][0]):
            yield doc


class TransactionManager(Manager):
    default_filter = "appl_id"

    def _get_results(self) -> tp.Iterator["Transaction"]:
        for doc in api.get_transactions(self.config.filter["appl_id"][0]):
            yield doc

    def count(self):
        return len(api.get_transactions(self.config.filter["appl_id"][0]))



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *           Source File: patent_client/_async/uspto/ptab/__init__.py           *
# ********************************************************************************

__api_name__ = "PTAB API v2"
from .model import PtabDecision, PtabDocument, PtabProceeding  # noqa: F401,E402

__all__ = [
    "PtabDecision",
    "PtabDocument",
    "PtabProceeding",
]



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *             Source File: patent_client/_async/uspto/odp/util.py              *
# ********************************************************************************

import typing as tp


def prune(obj: tp.Any) -> tp.Any:
    if isinstance(obj, dict):
        return {
            k: prune(v)
            for k, v in obj.items()
            if v is not None and not (isinstance(v, tp.Collection) and len(v) == 0)
        }
    elif isinstance(obj, list):
        return [
            prune(v)
            for v in obj
            if v is not None and not (isinstance(v, tp.Collection) and len(v) == 0)
        ]
    else:
        return obj



import datetime
import logging
import typing as tp
from collections.abc import Sequence
from pathlib import Path
from tempfile import TemporaryDirectory

from dateutil.parser import parse as dt_parse
from pypdf import PdfMerger

from patent_client.util.manager import AsyncManager
from patent_client.util.request_util import get_start_and_row_count

from .api import PatentExaminationDataSystemApi
from .query import QueryFields

if tp.TYPE_CHECKING:
    from .model import Document, USApplication

logger = logging.getLogger(__name__)


class HttpException(Exception):
    pass


def cast_as_datetime(
    date: tp.Union[str, datetime.datetime, datetime.date], end_of_day=False
) -> datetime.datetime:
    if isinstance(date, datetime.datetime):
        pass
    elif isinstance(date, str):
        date = dt_parse(date)
    elif isinstance(date, datetime.date):
        date = datetime.datetime.combine(date, datetime.datetime.min.time())
    elif not isinstance(date, datetime.datetime):
        raise ValueError(f"Invalid date type: {type(date)}")
    if end_of_day:
        date = date.replace(hour=23, minute=59, second=59)
    else:
        date = date.replace(hour=0, minute=0, second=0)
    return date


def datetime_to_solr(date: datetime.datetime) -> str:
    return date.strftime("%Y-%m-%dT%H:%M:%SZ")


class USApplicationManager(AsyncManager["USApplication"]):
    default_filter = "appl_id"

    async def count(self):
        api = PatentExaminationDataSystemApi()
        max_length = (await api.create_query(**self.get_query_params())).num_found
        return min(max_length, self.config.limit) if self.config.limit else max_length

    async def _get_results(self) -> tp.AsyncIterator["USApplication"]:
        query_params = self.get_query_params()
        api = PatentExaminationDataSystemApi()
        counter = 0
        for start, rows in get_start_and_row_count(
            self.config.limit, self.config.offset, page_size=20
        ):
            page = await api.create_query(**{**query_params, "start": start, "rows": rows})
            for app in page.applications:
                yield app
                counter += 1
                if self.config.limit and counter >= self.config.limit:
                    break
            if len(page.applications) < rows:
                break

    def get_query_params(self):
        # Short circuit processing logic if the "query" filter is specified
        if "query" in self.config.filter:
            query_text = self.config.filter["query"]
        else:
            query = list()
            date_filters = {
                QueryFields.get(k): v
                for k, v in self.config.filter.items()
                if QueryFields.is_date_field(k)
            }
            non_date_filters = {
                QueryFields.get(k): v
                for k, v in self.config.filter.items()
                if QueryFields.get(k) not in date_filters
            }

            date_filter_tuples = list()
            # Check date filters for validity
            for k, v in date_filters.items():
                if "__" in k:
                    f, *op = k.split("__")
                    if len(op) > 1:
                        raise ValueError(
                            f"Invalid date filter: {k}={v}; Cannot have more than one operator"
                        )
                    if op[0] not in ("gte", "lte"):
                        raise ValueError(f"Invalid date filter: {k}={v}; Invalid operator: {op[0]}")
                    if isinstance(v, (list, tuple)):
                        raise ValueError(
                            f"Invalid date filter: {k}={v}; Cannot have multiple values with operator {op[0]}"
                        )
                    date_filter_tuples.append((f, op[0], cast_as_datetime(v)))
                else:
                    if isinstance(v, (list, tuple)):
                        if len(v) != 2:
                            raise ValueError(
                                f"Invalid date range filter: {k}; When filtering using a list or tuple, must have length 2 (start, end)"
                            )
                        date_filter_tuples.append((k, "gte", cast_as_datetime(v[0])))
                        date_filter_tuples.append((k, "lte", cast_as_datetime(v[1])))
                    else:
                        date_filter_tuples.append((k, "exact", cast_as_datetime(v)))
            # Create pairs of gte/lte filters
            query_date_filter_tuples = set()
            for k, op, v in date_filter_tuples:
                if op == "gte":
                    lte_query = next(
                        (v for k, op, v in date_filter_tuples if k == k and op == "lte"),
                        "*",
                    )
                    query_date_filter_tuples.add((k, (v, lte_query)))
                elif op == "lte":
                    gte_query = next(
                        (v for k, op, v in date_filter_tuples if k == k and op == "gte"),
                        "*",
                    )
                    query_date_filter_tuples.add((k, (gte_query, v)))
                elif op == "exact":
                    query_date_filter_tuples.add((k, (v, v)))

            # Create the query string
            for k, v in query_date_filter_tuples:
                start, end = v
                start = datetime_to_solr(cast_as_datetime(start))
                end = datetime_to_solr(cast_as_datetime(end, end_of_day=True))
                query.append(f"{k}:[{start} TO {end}]")

            # Add non-date filters
            for k, v in non_date_filters.items():
                if isinstance(v, Sequence) and not isinstance(v, str):
                    values = " OR ".join(str(i) for i in v)
                    query.append(f"{k}:({values})")
                else:
                    query.append(f"{k}:({v})")
            query_text = " AND ".join(query)

        if self.config.order_by:
            sort_query = ""
            for s in self.config.order_by:
                if s[0] == "-":
                    sort_query += f"{QueryFields.get(s)} desc ".strip()
                else:
                    sort_query += (f"{QueryFields.get(s)} asc").strip()
        else:
            sort_query = None

        mm = "0%" if "appEarlyPubNumber" not in query else "90%"

        query_data = {
            "query": query_text,
            "facet": False,
            "minimum_match": mm,
        }
        if sort_query:
            query_data["sort"] = sort_query
        return query_data

    @property
    def allowed_filters(self):
        return QueryFields.field_names()

    async def is_online(self):
        return await PatentExaminationDataSystemApi.is_online()


class DocumentManager(AsyncManager["Document"]):
    default_filter = "appl_id"

    async def count(self):
        return len([i async for i in self])

    async def _get_results(self) -> tp.AsyncIterator["Document"]:
        api = PatentExaminationDataSystemApi()
        docs = await api.get_documents(self.config.filter["appl_id"][0])
        for doc in docs:
            yield doc

    async def download(self, docs, path="."):
        if str(path)[-4:].lower() == ".pdf":
            # If we've been given a specific filename, use it
            out_file = Path(path)
        else:
            out_file = Path(path) / "package.pdf"

        files = list()
        try:
            with TemporaryDirectory() as tmpdir:
                for doc in docs:
                    if doc.access_level_category == "PUBLIC":
                        files.append((await doc.adownload(tmpdir), doc))

                out_pdf = PdfMerger()
                page = 0
                for f, doc in files:
                    bookmark = f"{doc.mail_room_date} - {doc.code} - {doc.description}"
                    out_pdf.append(str(f), bookmark=bookmark, import_bookmarks=False)
                    page += doc.page_count

                out_pdf.write(str(out_file))
        except (PermissionError, NotADirectoryError):
            # This is needed due to a bug in Windows that prevents cleanup of the tmpdir
            pass
        return out_file



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *    Source File: patent_client/_async/epo/ops/published/schema/__init__.py    *
# ********************************************************************************

from .biblio import BiblioResultSchema
from .fulltext import ClaimsSchema, DescriptionSchema
from .images import ImagesSchema
from .search import SearchSchema

__all__ = [
    "BiblioResultSchema",
    "ClaimsSchema",
    "DescriptionSchema",
    "ImagesSchema",
    "SearchSchema",
]



import pytest

from .model.search import Inpadoc


class TestPublished:
    @pytest.mark.asyncio
    async def test_inpadoc_manager(self):
        result = Inpadoc.objects.filter(applicant="Microsoft")
        assert await result.count() > 20
        countries = [c async for c in result.limit(20).values_list("country", flat=True)]
        assert sum(1 for c in countries if c == "US") >= 1

    @pytest.mark.asyncio
    async def test_get_biblio_from_result(self):
        doc = await Inpadoc.objects.filter(applicant="Google").first()
        result = await doc.biblio
        assert len(result.titles) > 0

    @pytest.mark.asyncio
    async def test_get_claims_from_result(self):
        result = await Inpadoc.objects.get("WO2009085664A2")
        claims = await result.claims
        assert len(claims.claims) == 20
        assert len(claims.claim_text) == 4830

    @pytest.mark.asyncio
    async def test_get_description_from_result(self):
        result = await Inpadoc.objects.get("WO2009085664A2")
        description = await result.description
        assert len(description) == 47955

    @pytest.mark.asyncio
    async def test_get_family_from_result(self):
        result = await Inpadoc.objects.get("WO2009085664A2")
        family = await result.family
        assert len(family) >= 20

    @pytest.mark.asyncio
    async def test_get_biblio_from_wo(self):
        result = await Inpadoc.objects.get("WO2009085664A2")
        biblio = await result.biblio
        assert biblio.abstract is not None

    @pytest.mark.asyncio
    async def test_can_index_inpadoc_result(self):
        result = Inpadoc.objects.filter(applicant="Tesla")
        first = await result.first()
        second = await result.offset(1).first()
        assert first != second

    @pytest.mark.asyncio
    async def test_can_handle_single_item_ipc_classes(self):
        result = await Inpadoc.objects.get("WO2020081771")
        biblio = await result.biblio
        assert biblio.intl_class is not None

    @pytest.mark.asyncio
    async def test_issue_41(self):
        result = await Inpadoc.objects.get("JP2005533465A")
        biblio = await result.biblio
        assert biblio.title is None

    @pytest.mark.asyncio
    async def test_document_downloading(self, tmpdir):
        result = await Inpadoc.objects.get("WO2010050748A2")
        await result.download(tmpdir)
        assert len(tmpdir.listdir()) > 0



import datetime
import importlib
import re
import typing as tp

from async_property.base import AsyncPropertyDescriptor
from dateutil.parser import isoparse
from dateutil.parser import parse as parse_dt
from pydantic import BaseModel as PydanticBaseModel
from pydantic import BeforeValidator, ConfigDict
from typing_extensions import Annotated

from patent_client.util.manager import Manager


def parse_datetime(date_obj: tp.Union[str, datetime.datetime]) -> datetime.datetime:
    if isinstance(date_obj, datetime.datetime):
        return date_obj
    try:
        return isoparse(date_obj)
    except ValueError:
        return parse_dt(date_obj)


def parse_date(
    date_obj: tp.Union[str, datetime.datetime, datetime.date],
) -> datetime.date:
    if isinstance(date_obj, datetime.date):
        return date_obj
    elif isinstance(date_obj, datetime.datetime):
        return date_obj.date()
    try:
        return isoparse(date_obj).date()
    except ValueError:
        return parse_dt(date_obj).date()


DateTime = Annotated[datetime.datetime, BeforeValidator(parse_datetime)]
Date = Annotated[datetime.date, BeforeValidator(parse_date)]


class ClassProperty:
    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


class UnmanagedModelException(Exception):
    pass


M = tp.TypeVar("M", bound=Manager)


leading_period_re = re.compile(r"^\.+")


def get_class(class_name: str, base_class: type) -> type:
    if class_name.startswith("."):
        base_module = base_class.__module__
        # Count the number of dots to determine how many levels to go up
        level_up = leading_period_re.match(class_name).span()[1]
        module_path = base_module.split(".")[:-level_up]
        absolute_model_name = ".".join(module_path) + "." + class_name[level_up:]
    else:
        absolute_model_name = class_name

    module_name, class_name = absolute_model_name.rsplit(".", 1)
    try:
        module = importlib.import_module(module_name)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        raise ImportError(f"Could not import {absolute_model_name}: {e}")


class BaseModel(PydanticBaseModel, tp.Generic[M]):
    __manager__: tp.Optional[str] = None
    model_config = ConfigDict(
        ignored_types=(ClassProperty, AsyncPropertyDescriptor),
    )

    @ClassProperty
    def objects(cls) -> M:
        if cls.__manager__ is not None:
            manager_path = cls.__manager__
        else:
            manager_module = cls.__module__.split(".model")[0] + ".manager"
            manager_path = manager_module + "." + cls.__name__ + "Manager"
        return get_class(manager_path, base_class=cls)()

    def to_dict(self):
        return self.model_dump()

    def items(self):
        return iter(self)

    def _get_model(self, model_name: str, base_class=None) -> tp.Type["BaseModel"]:
        base_class = self.__class__ if base_class is None else base_class
        return get_class(model_name, base_class)



# ********************************************************************************
# *         WARNING: This file is automatically generated by unasync.py.         *
# *                             DO NOT MANUALLY EDIT                             *
# *          Source File: patent_client/_async/epo/ops/legal/manager.py          *
# ********************************************************************************

from yankee.data import ListCollection

from patent_client.util.manager import Manager

from .api import LegalApi
from .model import LegalEvent
from .schema import LegalSchema


class LegalManager(Manager):
    __schema__ = LegalSchema

    def get(self, doc_number, doc_type="publication", format="docdb") -> ListCollection[LegalEvent]:
        return self.__schema__.load(LegalApi.get_legal(doc_number, doc_type, format)).events



from typing import AsyncIterator, Generic, TypeVar

from patent_client.util.manager import AsyncManager
from patent_client.util.request_util import get_start_and_row_count

from .api import PublicSearchApi
from .model import (
    Patent,
    PatentBiblio,
    PublicSearchBiblio,
    PublicSearchDocument,
    PublishedApplication,
    PublishedApplicationBiblio,
)
from .query import QueryBuilder


class CapacityException(Exception):
    pass


class FinishedException(Exception):
    pass


T = TypeVar("T")

public_search_api = PublicSearchApi()


class GenericPublicSearchManager(AsyncManager, Generic[T]):
    page_size = 500
    default_filter = "patent_number"
    query_builder = QueryBuilder()

    @property
    def _query(self):
        return self.query_builder.build_query(self.config)

    @property
    def _order_by(self):
        return self.query_builder.build_order_by(self.config)

    @property
    def query_fields(self):
        return self.query_builder.search_keywords

    @property
    def order_by_fields(self):
        return self.query_builder.order_by_keywords

    async def count(self):
        query = self._query
        order_by = self._order_by
        sources = self.config.options.get("sources", ["US-PGPUB", "USPAT", "USOCR"])
        page = await public_search_api.run_query(
            query=query, start=0, limit=self.page_size, sort=order_by, sources=sources
        )
        max_len = page.num_found - self.config.offset
        return min(self.config.limit, max_len) if self.config.limit else max_len


class GenericPublicSearchBiblioManager(GenericPublicSearchManager, Generic[T]):
    async def _get_results(self) -> AsyncIterator[T]:
        query = self._query
        order_by = self._order_by
        sources = self.config.options.get("sources", ["US-PGPUB", "USPAT", "USOCR"])
        for start, rows in get_start_and_row_count(
            self.config.limit, self.config.offset, self.page_size
        ):
            page = await public_search_api.run_query(
                query=query,
                start=start,
                limit=rows,
                sort=order_by,
                sources=sources,
            )
            for obj in page.docs:
                yield obj
            if len(page.docs) < rows:
                break


capacity_limit = 501


class GenericPublicSearchDocumentManager(GenericPublicSearchBiblioManager, Generic[T]):
    async def _get_results(self) -> AsyncIterator["PublicSearchDocument"]:
        result_count = await super().count()
        if result_count > capacity_limit:
            raise CapacityException(
                f"Query would result in more than 501 results! ({result_count} > 20).\nPlease use the associated Biblio method to reduce load on the API (PublicSearch / PatentBiblio / PublishedApplicationBiblio"
            )
        async for obj in super()._get_results():
            doc = await public_search_api.get_document(obj)
            yield doc


class PublicSearchBiblioManager(GenericPublicSearchBiblioManager[PublicSearchBiblio]):
    pass


class PublicSearchDocumentManager(GenericPublicSearchDocumentManager[PublicSearchDocument]):
    pass


class PatentBiblioManager(GenericPublicSearchBiblioManager[PatentBiblio]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "USPAT",
        ]


class PatentManager(GenericPublicSearchDocumentManager[Patent]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "USPAT",
        ]


class PublishedApplicationBiblioManager(
    GenericPublicSearchBiblioManager[PublishedApplicationBiblio]
):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "US-PGPUB",
        ]


class PublishedApplicationManager(GenericPublicSearchDocumentManager[PublishedApplication]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "US-PGPUB",
        ]
